from __future__ import annotations

import asyncio
import warnings
from datetime import datetime, timedelta, timezone
from typing import TYPE_CHECKING

import pytest

from apify_client import ApifyClientAsync
from apify_shared.consts import ApifyEnvVars, WebhookEventType
from crawlee.events._types import Event

from apify import Actor, Webhook
from apify._actor import _ActorType

if TYPE_CHECKING:
    from ..conftest import ApifyClientAsyncPatcher


@pytest.fixture
def fake_actor_run() -> dict:
    return {
        'id': 'asdfasdf',
        'buildId': '3ads35',
        'buildNumber': '3.4.5',
        'actId': 'actor_id',
        'actorId': 'actor_id',
        'userId': 'user_id',
        'startedAt': '2024-08-08 12:12:44',
        'status': 'RUNNING',
        'meta': {'origin': 'API'},
        'containerUrl': 'http://0.0.0.0:3333',
        'defaultDatasetId': 'dhasdrfughaerguoi',
        'defaultKeyValueStoreId': 'asjkldhguiofg',
        'defaultRequestQueueId': 'lkjgklserjghios',
        'stats': {
            'inputBodyLen': 0,
            'restartCount': 0,
            'resurrectCount': 0,
            'memAvgBytes': 0,
            'memMaxBytes': 0,
            'memCurrentBytes': 0,
            'cpuAvgUsage': 0,
            'cpuMaxUsage': 0,
            'cpuCurrentUsage': 0,
            'netRxBytes': 0,
            'netTxBytes': 0,
            'durationMillis': 3333,
            'runTimeSecs': 33,
            'metamorph': 0,
            'computeUnits': 4.33,
        },
        'options': {
            'build': '',
            'timeoutSecs': 44,
            'memoryMbytes': 4096,
            'diskMbytes': 16384,
        },
    }


async def test_new_client_config_creation(monkeypatch: pytest.MonkeyPatch) -> None:
    token = 'my-token'
    monkeypatch.setenv(ApifyEnvVars.TOKEN, token)
    my_actor = _ActorType()
    await my_actor.init()

    client = my_actor.new_client()
    assert isinstance(client, ApifyClientAsync)
    assert client.token == token

    passed_token = 'my-passed-token'
    client_with_token = my_actor.new_client(token=passed_token)
    assert isinstance(client_with_token, ApifyClientAsync)
    assert client_with_token.token == passed_token

    await my_actor.exit()


async def test_call_actor(apify_client_async_patcher: ApifyClientAsyncPatcher, fake_actor_run: dict) -> None:
    apify_client_async_patcher.patch('actor', 'call', return_value=fake_actor_run)
    actor_id = 'some-actor-id'

    async with Actor:
        await Actor.call(actor_id)

    assert len(apify_client_async_patcher.calls['actor']['call']) == 1
    # The first argument is ActorClientAsync, which was called, let's check its id.
    assert apify_client_async_patcher.calls['actor']['call'][0][0][0].resource_id == actor_id


async def test_call_actor_task(apify_client_async_patcher: ApifyClientAsyncPatcher, fake_actor_run: dict) -> None:
    apify_client_async_patcher.patch('task', 'call', return_value=fake_actor_run)
    task_id = 'some-task-id'

    async with Actor:
        await Actor.call_task(task_id)

    assert len(apify_client_async_patcher.calls['task']['call']) == 1
    assert apify_client_async_patcher.calls['task']['call'][0][0][0].resource_id == task_id


async def test_start_actor(apify_client_async_patcher: ApifyClientAsyncPatcher, fake_actor_run: dict) -> None:
    apify_client_async_patcher.patch('actor', 'start', return_value=fake_actor_run)
    actor_id = 'some-id'

    async with Actor:
        await Actor.start(actor_id)

    assert len(apify_client_async_patcher.calls['actor']['start']) == 1
    assert apify_client_async_patcher.calls['actor']['start'][0][0][0].resource_id == actor_id


async def test_abort_actor_run(apify_client_async_patcher: ApifyClientAsyncPatcher, fake_actor_run: dict) -> None:
    apify_client_async_patcher.patch('run', 'abort', return_value=fake_actor_run)
    run_id = 'some-run-id'

    async with Actor:
        await Actor.abort(run_id)

    assert len(apify_client_async_patcher.calls['run']['abort']) == 1
    assert apify_client_async_patcher.calls['run']['abort'][0][0][0].resource_id == run_id


# NOTE: The following methods are properly tested using integrations tests.


async def test_metamorph_fails_locally(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level('WARNING')
    async with Actor:
        await Actor.metamorph('random-id')

    matching = [r for r in caplog.records if 'Actor.metamorph()' in r.message]
    assert len(matching) == 1
    assert matching[0].levelname == 'ERROR'
    assert 'only supported when running on the Apify platform' in matching[0].message


async def test_reboot_fails_locally(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level('WARNING')
    async with Actor:
        await Actor.reboot()

    matching = [r for r in caplog.records if 'Actor.reboot()' in r.message]
    assert len(matching) == 1
    assert matching[0].levelname == 'ERROR'
    assert 'only supported when running on the Apify platform' in matching[0].message


async def test_add_webhook_fails_locally(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level('WARNING')
    async with Actor:
        await Actor.add_webhook(
            Webhook(event_types=[WebhookEventType.ACTOR_BUILD_ABORTED], request_url='https://example.com')
        )

    matching = [r for r in caplog.records if 'Actor.add_webhook()' in r.message]
    assert len(matching) == 1
    assert matching[0].levelname == 'ERROR'
    assert 'only supported when running on the Apify platform' in matching[0].message


async def test_set_status_message_locally(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level('INFO')
    async with Actor:
        await Actor.set_status_message('test-status-message')

    matching_records = [record for record in caplog.records if 'test-status-message' in record.message]
    assert len(matching_records) == 1
    assert matching_records[0].levelname == 'INFO'
    assert '[Status message]: test-status-message' in matching_records[0].message


async def test_set_terminal_status_message_locally(caplog: pytest.LogCaptureFixture) -> None:
    caplog.set_level('INFO')
    async with Actor:
        await Actor.fail(status_message='test-terminal-message')

    matching_records = [record for record in caplog.records if 'test-terminal-message' in record.message]
    assert len(matching_records) == 1
    assert matching_records[0].levelname == 'INFO'
    assert '[Terminal status message]: test-terminal-message' in matching_records[0].message


async def test_push_data_with_empty_data() -> None:
    """Test that push_data returns result with `charged_count` 0 when data is empty."""
    async with Actor:
        result = await Actor.push_data([])
        assert result.charged_count == 0

        result = await Actor.push_data({})
        assert result.charged_count == 0


async def test_off_removes_event_listener(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that Actor.off() removes an event listener."""
    monkeypatch.setenv(ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS, '50')
    called = False

    async def listener(_data: object) -> None:
        nonlocal called
        called = True

    async with Actor:
        Actor.on(Event.PERSIST_STATE, listener)
        Actor.off(Event.PERSIST_STATE, listener)
        # Wait long enough for at least one PERSIST_STATE event to fire
        await asyncio.sleep(0.2)
        # Verify the listener was NOT called because it was removed
        assert called is False


_ACTOR_REMOTE_METHODS = [
    pytest.param('actor', 'start', 'start', 'some-actor-id', id='start'),
    pytest.param('actor', 'call', 'call', 'some-actor-id', id='call'),
    pytest.param('task', 'call', 'call_task', 'some-task-id', id='call_task'),
]


@pytest.mark.parametrize(('client_resource', 'client_method', 'actor_method_name', 'entity_id'), _ACTOR_REMOTE_METHODS)
async def test_remote_method_with_webhooks(
    apify_client_async_patcher: ApifyClientAsyncPatcher,
    fake_actor_run: dict,
    client_resource: str,
    client_method: str,
    actor_method_name: str,
    entity_id: str,
) -> None:
    """Test that start/call/call_task correctly serialize webhooks."""
    apify_client_async_patcher.patch(client_resource, client_method, return_value=fake_actor_run)

    async with Actor:
        actor_method = getattr(Actor, actor_method_name)
        await actor_method(
            entity_id,
            webhooks=[Webhook(event_types=[WebhookEventType.ACTOR_RUN_SUCCEEDED], request_url='https://example.com')],
        )

    calls = apify_client_async_patcher.calls[client_resource][client_method]
    assert len(calls) == 1
    _, kwargs = calls[0][0], calls[0][1]
    assert 'webhooks' in kwargs
    assert kwargs['webhooks'] is not None


@pytest.mark.parametrize(('client_resource', 'client_method', 'actor_method_name', 'entity_id'), _ACTOR_REMOTE_METHODS)
async def test_remote_method_with_timedelta_timeout(
    apify_client_async_patcher: ApifyClientAsyncPatcher,
    fake_actor_run: dict,
    client_resource: str,
    client_method: str,
    actor_method_name: str,
    entity_id: str,
) -> None:
    """Test that start/call/call_task accept a timedelta timeout."""
    apify_client_async_patcher.patch(client_resource, client_method, return_value=fake_actor_run)

    async with Actor:
        actor_method = getattr(Actor, actor_method_name)
        await actor_method(entity_id, timeout=timedelta(seconds=120))

    calls = apify_client_async_patcher.calls[client_resource][client_method]
    assert len(calls) == 1
    _, kwargs = calls[0][0], calls[0][1]
    assert kwargs.get('timeout_secs') == 120


async def test_call_actor_with_remaining_time_deprecation(
    apify_client_async_patcher: ApifyClientAsyncPatcher, fake_actor_run: dict
) -> None:
    """Test that call() with RemainingTime emits deprecation warning."""
    apify_client_async_patcher.patch('actor', 'call', return_value=fake_actor_run)

    async with Actor:
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always')
            await Actor.call('some-actor-id', timeout='RemainingTime')
            deprecation_warnings = [x for x in w if issubclass(x.category, DeprecationWarning)]
            assert len(deprecation_warnings) == 1
            assert 'RemainingTime' in str(deprecation_warnings[0].message)


@pytest.mark.parametrize(('client_resource', 'client_method', 'actor_method_name', 'entity_id'), _ACTOR_REMOTE_METHODS)
async def test_remote_method_with_invalid_timeout(
    apify_client_async_patcher: ApifyClientAsyncPatcher,
    fake_actor_run: dict,
    client_resource: str,
    client_method: str,
    actor_method_name: str,
    entity_id: str,
) -> None:
    """Test that start/call/call_task raise ValueError for invalid timeout."""
    apify_client_async_patcher.patch(client_resource, client_method, return_value=fake_actor_run)

    async with Actor:
        actor_method = getattr(Actor, actor_method_name)
        with pytest.raises(ValueError, match='Invalid timeout'):
            await actor_method(entity_id, timeout='invalid')


async def test_abort_with_status_message(
    apify_client_async_patcher: ApifyClientAsyncPatcher, fake_actor_run: dict
) -> None:
    """Test that abort() updates status message before aborting."""
    apify_client_async_patcher.patch('run', 'update', return_value=fake_actor_run)
    apify_client_async_patcher.patch('run', 'abort', return_value=fake_actor_run)

    async with Actor:
        await Actor.abort('run-id', status_message='Aborting due to error')

    assert len(apify_client_async_patcher.calls['run']['update']) == 1
    assert len(apify_client_async_patcher.calls['run']['abort']) == 1


async def test_get_remaining_time_warns_when_not_at_home(caplog: pytest.LogCaptureFixture) -> None:
    """Test that _get_remaining_time logs warning when not at home."""
    caplog.set_level('WARNING')
    async with Actor:
        # Actor is not at home, so _get_remaining_time should return None and log warning
        result = Actor._get_remaining_time()
        assert result is None
    assert any('inherit' in msg or 'RemainingTime' in msg for msg in caplog.messages)


async def test_get_remaining_time_clamps_negative_to_zero() -> None:
    """Test that _get_remaining_time returns timedelta(0) instead of a negative value when timeout is in the past."""
    async with Actor:
        Actor.configuration.is_at_home = True
        Actor.configuration.timeout_at = datetime.now(tz=timezone.utc) - timedelta(minutes=5)

        result = Actor._get_remaining_time()
        assert result is not None
        assert result == timedelta(0)


async def test_get_remaining_time_returns_positive_when_timeout_in_future() -> None:
    """Test that _get_remaining_time returns a positive timedelta when timeout is in the future."""
    async with Actor:
        Actor.configuration.is_at_home = True
        Actor.configuration.timeout_at = datetime.now(tz=timezone.utc) + timedelta(minutes=5)

        result = Actor._get_remaining_time()
        assert result is not None
        assert result > timedelta(0)
        assert result <= timedelta(minutes=5)



import asyncio
import random

from apify import Actor


async def do_work() -> None:
    # Simulate random outcomes: success or one of two exception types.
    outcome = random.random()

    if outcome < 0.33:
        raise ValueError('Invalid input data encountered')
    if outcome < 0.66:
        raise RuntimeError('Unexpected runtime failure')

    # Simulate successful work
    Actor.log.info('Work completed successfully')


async def main() -> None:
    await Actor.init()
    try:
        await do_work()
    except ValueError as exc:  # Specific error mapping example
        await Actor.fail(exit_code=10, exception=exc)
    except Exception as exc:  # Catch-all for unexpected errors
        await Actor.fail(exit_code=91, exception=exc)
    else:
        await Actor.exit(status_message='Actor completed successfully')


if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations

from base64 import b64encode
from typing import TYPE_CHECKING
from urllib.parse import unquote

from scrapy.utils.project import get_project_settings
from scrapy.utils.python import to_bytes

if TYPE_CHECKING:
    from scrapy.settings import Settings


def get_basic_auth_header(username: str, password: str, auth_encoding: str = 'latin-1') -> bytes:
    """Generate a basic authentication header for the given username and password."""
    string = f'{unquote(username)}:{unquote(password)}'
    user_pass = to_bytes(string, encoding=auth_encoding)
    return b'Basic ' + b64encode(user_pass)


def apply_apify_settings(*, settings: Settings | None = None, proxy_config: dict | None = None) -> Settings:
    """Integrates Apify configuration into a Scrapy project settings.

    Note: The function directly modifies the passed `settings` object and also returns it.

    Args:
        settings: Scrapy project settings to be modified.
        proxy_config: Proxy configuration to be stored in the settings.

    Returns:
        Scrapy project settings with custom configurations.
    """
    if settings is None:
        settings = get_project_settings()

    # Use ApifyScheduler as the scheduler
    settings['SCHEDULER'] = 'apify.scrapy.scheduler.ApifyScheduler'

    # Add the ActorDatasetPushPipeline into the item pipelines, assigning it the highest integer (1000),
    # ensuring it is executed as the final step in the pipeline sequence
    settings['ITEM_PIPELINES']['apify.scrapy.pipelines.ActorDatasetPushPipeline'] = 1000

    # Replace the default HttpProxyMiddleware with ApifyHttpProxyMiddleware
    settings['DOWNLOADER_MIDDLEWARES']['scrapy.downloadermiddlewares.httpproxy.HttpProxyMiddleware'] = None
    settings['DOWNLOADER_MIDDLEWARES']['apify.scrapy.middlewares.ApifyHttpProxyMiddleware'] = 750

    # Set the default HTTPCache middleware storage backend to ApifyCacheStorage
    settings['HTTPCACHE_STORAGE'] = 'apify.scrapy.extensions.ApifyCacheStorage'

    # Store the proxy configuration
    settings['APIFY_PROXY_SETTINGS'] = proxy_config

    return settings



from __future__ import annotations

import ipaddress
import json
import re
from dataclasses import dataclass, field
from re import Pattern
from typing import TYPE_CHECKING, Any
from urllib.parse import urljoin, urlparse

import impit
from yarl import URL

from apify_shared.consts import ApifyEnvVars
from crawlee.proxy_configuration import ProxyConfiguration as CrawleeProxyConfiguration
from crawlee.proxy_configuration import ProxyInfo as CrawleeProxyInfo
from crawlee.proxy_configuration import _NewUrlFunction

from apify._configuration import Configuration
from apify._utils import docs_group
from apify.log import logger

if TYPE_CHECKING:
    from apify_client import ApifyClientAsync

    from apify import Request

APIFY_PROXY_VALUE_REGEX = re.compile(r'^[\w._~]+$')
COUNTRY_CODE_REGEX = re.compile(r'^[A-Z]{2}$')
SESSION_ID_MAX_LENGTH = 50


def is_url(url: str) -> bool:
    """Check if the given string is a valid URL."""
    try:
        parsed_url = urlparse(urljoin(url, '/'))
        has_all_parts = all([parsed_url.scheme, parsed_url.netloc, parsed_url.path])
        is_domain = '.' in parsed_url.netloc
        is_localhost = parsed_url.netloc == 'localhost'
        try:
            ipaddress.ip_address(parsed_url.netloc)
            is_ip_address = True
        except Exception:
            is_ip_address = False

        return has_all_parts and any([is_domain, is_localhost, is_ip_address])
    except Exception:
        return False


def _check(
    value: Any,
    *,
    label: str | None,
    pattern: Pattern | None = None,
    min_length: int | None = None,
    max_length: int | None = None,
) -> None:
    error_str = f'Value {value}'
    if label:
        error_str += f' of argument {label}'

    if min_length and len(value) < min_length:
        raise ValueError(f'{error_str} is shorter than minimum allowed length {min_length}')

    if max_length and len(value) > max_length:
        raise ValueError(f'{error_str} is longer than maximum allowed length {max_length}')

    if pattern and not re.fullmatch(pattern, value):
        raise ValueError(f'{error_str} does not match pattern {pattern.pattern!r}')


@docs_group('Configuration')
@dataclass
class ProxyInfo(CrawleeProxyInfo):
    """Provides information about a proxy connection that is used for requests."""

    groups: list[str] = field(default_factory=list)
    """An array of proxy groups to be used by the [Apify Proxy](https://docs.apify.com/proxy). If not provided,
    the proxy will select the groups automatically."""

    country_code: str | None = None
    """If set and relevant proxies are available in your Apify account, all proxied requests will use IP addresses
    that are geolocated to the specified country. For example `GB` for IPs from Great Britain. Note that online
    services often have their own rules for handling geolocation and thus the country selection is a best attempt
    at geolocation, rather than a guaranteed hit. This parameter is optional, by default, each proxied request is
    assigned an IP address from a random country. The country code needs to be a two letter ISO country code.
    See the [full list of available country codes](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements).
    This parameter is optional, by default, the proxy uses all available proxy servers from all countries.
    """


@docs_group('Configuration')
class ProxyConfiguration(CrawleeProxyConfiguration):
    """Configures a connection to a proxy server with the provided options.

    Proxy servers are used to prevent target websites from blocking your crawlers based on IP address rate limits or
    blacklists. The default servers used by this class are managed by [Apify Proxy](https://docs.apify.com/proxy).
    To be able to use Apify Proxy, you need an Apify account and access to the selected proxies. If you provide
    no configuration option, the proxies will be managed automatically using a smart algorithm.

    If you want to use your own proxies, use the `proxy_urls` or `new_url_function` constructor options. Your list
    of proxy URLs will be rotated by the configuration, if this option is provided.
    """

    _configuration: Configuration

    def __init__(
        self,
        *,
        password: str | None = None,
        groups: list[str] | None = None,
        country_code: str | None = None,
        proxy_urls: list[str | None] | None = None,
        new_url_function: _NewUrlFunction | None = None,
        tiered_proxy_urls: list[list[str | None]] | None = None,
        _actor_config: Configuration | None = None,
        _apify_client: ApifyClientAsync | None = None,
    ) -> None:
        """Create a ProxyConfiguration instance.

        It is highly recommended to use `Actor.create_proxy_configuration()` instead of this.

        Args:
            password: Password for the Apify Proxy. If not provided, will use os.environ['APIFY_PROXY_PASSWORD'],
                if available.
            groups: Proxy groups which the Apify Proxy should use, if provided.
            country_code: Country which the Apify Proxy should use, if provided.
            proxy_urls: Custom proxy server URLs which should be rotated through.
            new_url_function: Function which returns a custom proxy URL to be used.
            tiered_proxy_urls: Proxy URLs arranged into tiers
        """
        _actor_config = _actor_config or Configuration.get_global_configuration()

        if groups:
            groups = [str(group) for group in groups]
            for group in groups:
                _check(group, label='groups', pattern=APIFY_PROXY_VALUE_REGEX)

        if country_code:
            country_code = str(country_code)
            _check(country_code, label='country_code', pattern=COUNTRY_CODE_REGEX)

        if (proxy_urls or new_url_function or tiered_proxy_urls) and (groups or country_code):
            raise ValueError(
                'Cannot combine custom proxies with Apify Proxy!'
                ' It is not allowed to set "proxy_urls" or "new_url_function" combined with'
                ' "groups" or "country_code".'
            )

        if proxy_urls and any('apify.com' in (url or '') for url in proxy_urls):
            logger.warning(
                'Some Apify proxy features may work incorrectly. Please consider setting up Apify properties '
                'instead of `proxy_urls`.\n'
                'See https://sdk.apify.com/docs/guides/proxy-management#apify-proxy-configuration'
            )

        self._uses_apify_proxy = not (proxy_urls or new_url_function or tiered_proxy_urls)

        super().__init__(
            proxy_urls=[f'http://{_actor_config.proxy_hostname}:{_actor_config.proxy_port}']
            if self._uses_apify_proxy
            else proxy_urls,
            new_url_function=new_url_function,
            tiered_proxy_urls=tiered_proxy_urls,
        )
        self._configuration = _actor_config

        self.is_man_in_the_middle = False

        self._apify_client = _apify_client

        self._hostname = self._configuration.proxy_hostname
        self._port = self._configuration.proxy_port
        self._password = password or self._configuration.proxy_password

        self._groups = list(groups) if groups else []
        self._country_code = country_code

    async def initialize(self) -> None:
        """Check if using proxy, if so, check the access.

        Load the Apify Proxy password from API (only if not passed to constructor or through env var).

        Only called if Apify Proxy configuration is used. Also checks if country has access to Apify Proxy groups
        if the country code is provided.

        You should use the Actor.create_proxy_configuration function to create a pre-initialized
        `ProxyConfiguration` instance instead of calling this manually.
        """
        if self._uses_apify_proxy:
            if not self._password:
                await self._maybe_fetch_password()
            if not self._password:
                raise ValueError(
                    'Apify Proxy password must be provided using the "password" constructor argument '
                    f'or the "{ApifyEnvVars.PROXY_PASSWORD}" environment variable. '
                    f'You can also provide your Apify token via the "${ApifyEnvVars.TOKEN}" environment variable, '
                    f'so that the SDK can fetch the proxy password from Apify API, '
                    f'when not provided through constructor or ${ApifyEnvVars.PROXY_PASSWORD}.'
                )

            await self._check_access()

    async def new_proxy_info(
        self,
        session_id: str | None = None,
        request: Request | None = None,
        proxy_tier: int | None = None,
    ) -> ProxyInfo | None:
        """Create a new ProxyInfo object.

        Use it if you want to work with a rich representation of a proxy URL. If you need the URL string only,
        use `ProxyConfiguration.new_url`.

        Args:
            session_id: Represents the identifier of a proxy session (https://docs.apify.com/proxy#sessions).
                All the HTTP requests going through the proxy with the same session identifier will use the same
                target proxy server (i.e. the same IP address). The identifier must not be longer than 50 characters
                and include only the following: `0-9`, `a-z`, `A-Z`, `"."`, `"_"` and `"~"`.
            request: request for which the proxy info is being issued, used in proxy tier handling.
            proxy_tier: allows forcing the proxy tier to be used.

        Returns:
            Dictionary that represents information about the proxy and its configuration.
        """
        if session_id is not None:
            _check(session_id, label='session_id', max_length=SESSION_ID_MAX_LENGTH, pattern=APIFY_PROXY_VALUE_REGEX)

        proxy_info = await super().new_proxy_info(session_id=session_id, request=request, proxy_tier=proxy_tier)

        if proxy_info is None:
            return None

        if self._uses_apify_proxy:
            parsed_url = URL(proxy_info.url)
            username = self._get_username(session_id)

            return ProxyInfo(
                url=f'http://{username}:{self._password or ""}@{parsed_url.host}:{parsed_url.port}',
                scheme='http',
                hostname=proxy_info.hostname,
                port=proxy_info.port,
                username=username,
                password=self._password or '',
                session_id=proxy_info.session_id,
                proxy_tier=proxy_info.proxy_tier,
                groups=self._groups,
                country_code=self._country_code or None,
            )

        return ProxyInfo(
            url=proxy_info.url,
            scheme=proxy_info.scheme,
            hostname=proxy_info.hostname,
            port=proxy_info.port,
            username=proxy_info.username,
            password=proxy_info.password,
            session_id=proxy_info.session_id,
            proxy_tier=proxy_info.proxy_tier,
        )

    async def _maybe_fetch_password(self) -> None:
        token = self._configuration.token

        if token and self._apify_client:
            user_info = await self._apify_client.user().get()
            if user_info:
                password = user_info['proxy']['password']
                self._password = password

    async def _check_access(self) -> None:
        proxy_status_url = f'{self._configuration.proxy_status_url}/?format=json'
        proxy_info = await self.new_proxy_info()

        if proxy_info is None:
            return

        status = None
        async with impit.AsyncClient(proxy=proxy_info.url, timeout=10) as client:
            for _ in range(2):
                try:
                    response = await client.get(proxy_status_url)
                    status = json.loads(response.text)
                    break
                except Exception:  # noqa: S110
                    # retry on connection errors
                    pass

        if status:
            if not status['connected']:
                raise ConnectionError(status['connectionError'])

            self.is_man_in_the_middle = status['isManInTheMiddle']
        else:
            logger.warning(
                'Apify Proxy access check timed out. Watch out for errors with status code 407. If you see some, it '
                'most likely means you do not have access to either all or some of the proxies you are trying to use.'
            )

    def _get_username(self, session_id: int | str | None = None) -> str:
        if session_id is not None:
            session_id = f'{session_id}'

        parts: list[str] = []

        if self._groups:
            parts.append(f'groups-{"+".join(self._groups)}')
        if session_id is not None:
            parts.append(f'session-{session_id}')
        if self._country_code:
            parts.append(f'country-{self._country_code}')

        if not parts:
            return 'auto'

        return ','.join(parts)



import asyncio
from datetime import timedelta

from apify import Actor


async def main() -> None:
    actor = Actor(
        event_listeners_timeout=timedelta(seconds=30),
        cleanup_timeout=timedelta(seconds=30),
    )

    await actor.init()

    try:
        # Get input
        actor_input = await actor.get_input()
        actor.log.info('Actor input: %s', actor_input)

        # Your Actor logic here
        data = {'message': 'Hello from Actor!', 'input': actor_input}
        await actor.push_data(data)

        # Set status message
        await actor.set_status_message('Actor completed successfully')

    finally:
        await actor.exit()


if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations

import random
import string
from datetime import datetime, timedelta
from decimal import Decimal
from typing import TYPE_CHECKING, Any

from pydantic_core import TzInfo

from apify_shared.consts import (
    BOOL_ENV_VARS,
    COMMA_SEPARATED_LIST_ENV_VARS,
    DATETIME_ENV_VARS,
    FLOAT_ENV_VARS,
    INTEGER_ENV_VARS,
    STRING_ENV_VARS,
    ActorEnvVars,
    ApifyEnvVars,
)

from apify import Actor

if TYPE_CHECKING:
    from pathlib import Path

    import pytest


async def test_actor_is_not_at_home_when_local() -> None:
    async with Actor as actor:
        is_at_home = actor.is_at_home()
        assert is_at_home is False


async def test_get_env_with_randomized_env_vars(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    ignored_env_vars = {
        ApifyEnvVars.SDK_LATEST_VERSION,
        ApifyEnvVars.LOG_FORMAT,
        ApifyEnvVars.LOG_LEVEL,
        ApifyEnvVars.LOCAL_STORAGE_DIR,
        ActorEnvVars.STANDBY_PORT,
        ApifyEnvVars.PERSIST_STORAGE,
    }

    # Set up random env vars
    expected_get_env = dict[str, Any]()
    expected_get_env[ApifyEnvVars.LOG_LEVEL.name.lower()] = 'INFO'

    for int_env_var in INTEGER_ENV_VARS:
        if int_env_var in ignored_env_vars:
            continue

        int_get_env_var = int_env_var.name.lower()
        expected_get_env[int_get_env_var] = random.randint(1, 99999)
        monkeypatch.setenv(int_env_var, f'{expected_get_env[int_get_env_var]}')

    for float_env_var in FLOAT_ENV_VARS:
        if float_env_var in ignored_env_vars:
            continue

        float_get_env_var = float_env_var.name.lower()
        if float_env_var == ActorEnvVars.MAX_TOTAL_CHARGE_USD:
            expected_get_env[float_get_env_var] = Decimal(random.random())
        else:
            expected_get_env[float_get_env_var] = random.random()
        monkeypatch.setenv(float_env_var, f'{expected_get_env[float_get_env_var]}')

    for bool_env_var in BOOL_ENV_VARS:
        if bool_env_var in ignored_env_vars:
            continue

        bool_get_env_var = bool_env_var.name.lower()
        expected_get_env[bool_get_env_var] = random.choice([True, False])
        monkeypatch.setenv(bool_env_var, f'{"true" if expected_get_env[bool_get_env_var] else "false"}')

    expected_get_env[ApifyEnvVars.IS_AT_HOME.name.lower()] = False
    monkeypatch.setenv(ApifyEnvVars.IS_AT_HOME, 'false')

    for datetime_env_var in DATETIME_ENV_VARS:
        if datetime_env_var in ignored_env_vars:
            continue

        datetime_get_env_var = datetime_env_var.name.lower()
        expected_get_env[datetime_get_env_var] = datetime.now(TzInfo(0))
        monkeypatch.setenv(
            datetime_env_var,
            expected_get_env[datetime_get_env_var].strftime('%Y-%m-%dT%H:%M:%S.%fZ'),
        )

    for string_env_var in STRING_ENV_VARS:
        if string_env_var in ignored_env_vars:
            continue

        string_get_env_var = string_env_var.name.lower()
        expected_value = ''.join(random.choices(string.ascii_uppercase + string.digits, k=10))
        # URLs have to be valid
        if string_get_env_var.endswith('url'):
            expected_value = f'http://example.com/{expected_value}'
        expected_get_env[string_get_env_var] = expected_value
        monkeypatch.setenv(string_env_var, expected_get_env[string_get_env_var])

    for list_env_var in COMMA_SEPARATED_LIST_ENV_VARS:
        if list_env_var in ignored_env_vars:
            continue

        available_values = ['val1', 'val2']

        list_get_env_var = list_env_var.name.lower()
        expected_value_count = random.randint(0, len(available_values))
        expected_get_env[list_get_env_var] = random.sample(available_values, expected_value_count)
        monkeypatch.setenv(list_env_var, ','.join(expected_get_env[list_get_env_var]))

        # Test behavior with missing env var in case of empty list
        if expected_value_count == 0 and random.random() < 0.5:
            monkeypatch.delenv(list_env_var)
            expected_get_env[list_get_env_var] = None

    # We need this override so that the actor doesn't fail when connecting to the platform events websocket
    monkeypatch.delenv(ActorEnvVars.EVENTS_WEBSOCKET_URL)
    expected_get_env[ActorEnvVars.EVENTS_WEBSOCKET_URL.name.lower()] = None

    # Adjust expectations for timedelta fields
    for env_name, env_value in expected_get_env.items():
        if env_name.endswith('_millis'):
            expected_get_env[env_name] = timedelta(milliseconds=env_value)

    # Convert dedicated_cpus to float
    expected_get_env[ApifyEnvVars.DEDICATED_CPUS.name.lower()] = float(
        expected_get_env[ApifyEnvVars.DEDICATED_CPUS.name.lower()]
    )

    # LOCAL_STORAGE_DIR is excluded from randomization to avoid creating directories in the project root.
    # Its value is set by the _isolate_test_environment fixture to tmp_path.
    expected_get_env[ApifyEnvVars.LOCAL_STORAGE_DIR.name.lower()] = str(tmp_path)

    await Actor.init()
    assert Actor.get_env() == expected_get_env

    await Actor.exit()



from __future__ import annotations

import re

from crawlee.crawlers import HttpCrawler, HttpCrawlingContext

from apify import Actor


async def main() -> None:
    async with Actor:
        pages_visited: list[str] = []
        crawler = HttpCrawler(max_crawl_depth=2)

        @crawler.router.default_handler
        async def handler(context: HttpCrawlingContext) -> None:
            pages_visited.append(context.request.url)
            html = (await context.http_response.read()).decode()

            links = re.findall(r'href="(/[^"]*)"', html)
            base_url = 'http://localhost:8080'
            await context.add_requests([f'{base_url}{link}' for link in links])

            if '/products/' in context.request.url:
                name_match = re.search(r'<h1>(.*?)</h1>', html)
                price_match = re.search(r'<span class="price">(.*?)</span>', html)
                desc_match = re.search(r'<p class="description">(.*?)</p>', html)
                if name_match:
                    await context.push_data(
                        {
                            'url': context.request.url,
                            'name': name_match.group(1),
                            'price': price_match.group(1) if price_match else '',
                            'description': desc_match.group(1) if desc_match else '',
                        }
                    )

        await crawler.run(['http://localhost:8080/'])

        await Actor.set_value(
            'CRAWLER_RESULT',
            {
                'pages_visited_count': len(pages_visited),
                'crawler_type': 'HttpCrawler',
            },
        )



from __future__ import annotations

import pytest

from apify import Actor


async def test_open_throws_without_init() -> None:
    with pytest.raises(RuntimeError):
        await Actor.open_request_queue()


async def test_open_returns_same_references() -> None:
    async with Actor:
        rq1 = await Actor.open_request_queue()
        rq2 = await Actor.open_request_queue()
        assert rq1 is rq2

        rq_name = 'non-default'
        rq_by_name_1 = await Actor.open_request_queue(name=rq_name)
        rq_by_name_2 = await Actor.open_request_queue(name=rq_name)
        assert rq_by_name_1 is rq_by_name_2

        rq_1_metadata = await rq_by_name_1.get_metadata()
        rq_by_id_1 = await Actor.open_request_queue(id=rq_1_metadata.id)
        rq_by_id_2 = await Actor.open_request_queue(id=rq_1_metadata.id)
        assert rq_by_id_1 is rq_by_name_1
        assert rq_by_id_2 is rq_by_id_1



from apify import Actor


async def main() -> None:
    async with Actor:
        # Open a key-value store with the name 'my-cool-store'
        key_value_store = await Actor.open_key_value_store(name='my-cool-store')
        await key_value_store.set_value('record', 'Hello, world!')

        # Do something ...

        # Now we don't want it anymore
        await key_value_store.drop()



from __future__ import annotations

from html.parser import HTMLParser
from typing import TYPE_CHECKING

from crawlee.crawlers import BasicCrawler

from apify import Actor

if TYPE_CHECKING:
    from crawlee._types import BasicCrawlingContext


class _PageParser(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self.links: list[str] = []
        self.data: dict[str, str] = {}
        self._in_tag: str | None = None
        self._in_class: str = ''

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:
        attrs_dict = dict(attrs)
        if tag == 'a' and (href := attrs_dict.get('href')):
            self.links.append(href)
        self._in_tag = tag
        self._in_class = attrs_dict.get('class', '') or ''

    def handle_endtag(self, tag: str) -> None:  # noqa: ARG002
        self._in_tag = None
        self._in_class = ''

    def handle_data(self, data: str) -> None:
        text = data.strip()
        if not text:
            return
        if self._in_tag == 'h1':
            self.data['name'] = text
        elif self._in_tag == 'span' and self._in_class == 'price':
            self.data['price'] = text
        elif self._in_tag == 'p' and self._in_class == 'description':
            self.data['description'] = text


async def main() -> None:
    async with Actor:
        pages_visited: list[str] = []
        crawler = BasicCrawler(max_crawl_depth=2)

        @crawler.router.default_handler
        async def handler(context: BasicCrawlingContext) -> None:
            pages_visited.append(context.request.url)

            response = await context.send_request(context.request.url)
            html = (await response.read()).decode()

            parser = _PageParser()
            parser.feed(html)

            base_url = 'http://localhost:8080'
            await context.add_requests([f'{base_url}{link}' for link in parser.links if link.startswith('/')])

            if '/products/' in context.request.url and parser.data.get('name'):
                await context.push_data({'url': context.request.url, **parser.data})

        await crawler.run(['http://localhost:8080/'])

        await Actor.set_value(
            'CRAWLER_RESULT',
            {
                'pages_visited_count': len(pages_visited),
                'crawler_type': 'BasicCrawler',
            },
        )



from __future__ import annotations

from datetime import timedelta

from apify import Actor, Configuration


async def test_actor_with_non_default_config() -> None:
    config = Configuration(internal_timeout=timedelta(minutes=111))

    async with Actor(config) as actor:
        assert actor.configuration.internal_timeout == timedelta(minutes=111)


async def test_actor_global_works() -> None:
    non_default_configuration = Configuration(actor_full_name='Actor McActorson, esq.')

    async with Actor(configuration=non_default_configuration):
        assert Actor.configuration is non_default_configuration



from apify import Actor


async def main() -> None:
    async with Actor:
        proxy_configuration = await Actor.create_proxy_configuration()

        if not proxy_configuration:
            raise RuntimeError('No proxy configuration available.')

        proxy_url = await proxy_configuration.new_url()
        Actor.log.info(f'Using proxy URL: {proxy_url}')



from time import time

import pytest

from apify.scrapy.extensions._httpcache import from_gzip, get_kvs_name, read_gzip_time, to_gzip

FIXTURE_DICT = {'name': 'Alice'}

FIXTURE_BYTES = (
    b'\x1f\x8b\x08\x00\x00\x00\x00\x00\x02\xffk`\x99*\xcc\x00\x01\xb5SzX\xf2\x12s'
    b'S\xa7\xf4\xb0:\xe6d&\xa7N)\xd6\x03\x00\x1c\xe8U\x9c\x1e\x00\x00\x00'
)


def test_gzip() -> None:
    assert from_gzip(to_gzip(FIXTURE_DICT)) == FIXTURE_DICT


def test_to_gzip() -> None:
    data_bytes = to_gzip(FIXTURE_DICT, mtime=0)

    assert data_bytes == FIXTURE_BYTES


def test_from_gzip() -> None:
    data_dict = from_gzip(FIXTURE_BYTES)

    assert data_dict == FIXTURE_DICT


def test_read_gzip_time() -> None:
    assert read_gzip_time(FIXTURE_BYTES) == 0


def test_read_gzip_time_non_zero() -> None:
    current_time = int(time())
    data_bytes = to_gzip(FIXTURE_DICT, mtime=current_time)

    assert read_gzip_time(data_bytes) == current_time


@pytest.mark.parametrize(
    ('spider_name', 'expected'),
    [
        ('test', 'httpcache-test'),
        ('123', 'httpcache-123'),
        ('test-spider', 'httpcache-test-spider'),
        ('test_spider', 'httpcache-test-spider'),
        ('test spider', 'httpcache-test-spider'),
        ('test👻spider', 'httpcache-test-spider'),
        ('test@spider', 'httpcache-test-spider'),
        ('   test   spider   ', 'httpcache-test-spider'),
        ('testspider.com', 'httpcache-testspider-com'),
        ('t' * 100, 'httpcache-tttttttttttttttttttttttttttttttttttttttttttttttttt'),
    ],
)
def test_get_kvs_name(spider_name: str, expected: str) -> None:
    assert get_kvs_name(spider_name) == expected


@pytest.mark.parametrize(
    ('spider_name'),
    [
        '',
        '-',
        '-@-/-',
    ],
)
def test_get_kvs_name_raises(spider_name: str) -> None:
    with pytest.raises(ValueError, match=r'Unsupported spider name'):
        assert get_kvs_name(spider_name)



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

from apify_shared.consts import ActorEventTypes

from apify import Actor

if TYPE_CHECKING:
    from .conftest import MakeActorFunction, RunActorFunction


async def test_emit_and_capture_interval_events(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        # TC003 - TYPE_CHECKING section is problematic for our integration tests.
        import os
        from collections.abc import Callable  # noqa: TC003
        from datetime import datetime
        from typing import Any

        from apify_shared.consts import ActorEventTypes, ApifyEnvVars
        from crawlee.events._types import Event, EventSystemInfoData

        os.environ[ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS] = '900'

        was_system_info_emitted = False
        system_infos = list[EventSystemInfoData]()

        def on_event(event_type: ActorEventTypes) -> Callable:
            async def log_event(data: Any) -> None:
                nonlocal was_system_info_emitted
                nonlocal system_infos
                print(f'Got actor event ({event_type=}, {data=})')
                await Actor.push_data({'event_type': event_type, 'data': data})
                if event_type == ActorEventTypes.SYSTEM_INFO:
                    was_system_info_emitted = True
                    system_infos.append(data)

            return log_event

        async with Actor:
            Actor.on(Event.SYSTEM_INFO, on_event(ActorEventTypes.SYSTEM_INFO))
            Actor.on(Event.PERSIST_STATE, on_event(ActorEventTypes.PERSIST_STATE))
            await asyncio.sleep(3)

            # The SYSTEM_INFO event sometimes takes a while to appear, let's wait for it for a while longer.
            for _ in range(20):
                if was_system_info_emitted:
                    break
                await asyncio.sleep(1)

            # Check that parsing datetimes works correctly.
            # Check `createdAt` is a datetime (so it's the same locally and on platform).
            assert isinstance(system_infos[0].cpu_info.created_at, datetime)

    actor = await make_actor(label='actor-interval-events', main_func=main)
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'

    dataset_items_page = await actor.last_run().dataset().list_items()
    persist_state_events = [
        item for item in dataset_items_page.items if item['event_type'] == ActorEventTypes.PERSIST_STATE
    ]
    system_info_events = [
        item for item in dataset_items_page.items if item['event_type'] == ActorEventTypes.SYSTEM_INFO
    ]
    assert len(persist_state_events) > 2
    assert len(system_info_events) > 0



from __future__ import annotations

from apify import Actor


async def main() -> None:
    async with Actor:
        raise RuntimeError('You need to override the `main.py` file in the integration test!')



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..conftest import MakeActorFunction, RunActorFunction

_ACTOR_SOURCE_DIR = Path(__file__).parent / 'actor_source'


def read_actor_source(filename: str) -> str:
    return (_ACTOR_SOURCE_DIR / filename).read_text()


async def test_configuration_storages(make_actor: MakeActorFunction, run_actor: RunActorFunction) -> None:
    actor = await make_actor(
        label='schema_storages',
        source_files={
            'src/main.py': read_actor_source('main.py'),
            '.actor/actor.json': read_actor_source('actor.json'),
        },
    )
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'



from __future__ import annotations

import base64
import hashlib
import hmac
import json
import string
from typing import Any

from cryptography.exceptions import InvalidTag as InvalidTagException
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric import padding, rsa
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from crawlee._utils.crypto import crypto_random_object_id

from apify._consts import ENCRYPTED_INPUT_VALUE_REGEXP, ENCRYPTED_JSON_VALUE_PREFIX, ENCRYPTED_STRING_VALUE_PREFIX

ENCRYPTION_KEY_LENGTH = 32
ENCRYPTION_IV_LENGTH = 16
ENCRYPTION_AUTH_TAG_LENGTH = 16


def public_encrypt(value: str, *, public_key: rsa.RSAPublicKey) -> dict:
    """Encrypts the given value using AES cipher and the password for encryption using the public key.

    The encryption password is a string of encryption key and initial vector used for cipher.
    It returns the encrypted password and encrypted value in BASE64 format.

    Args:
        value: The value which should be encrypted.
        public_key: Public key to use for encryption.

    Returns: Encrypted password and value.
    """
    key_bytes = crypto_random_object_id(ENCRYPTION_KEY_LENGTH).encode('utf-8')
    initialized_vector_bytes = crypto_random_object_id(ENCRYPTION_IV_LENGTH).encode('utf-8')
    value_bytes = value.encode('utf-8')

    password_bytes = key_bytes + initialized_vector_bytes

    # NOTE: Auth Tag is appended to the end of the encrypted data, it has length of 16 bytes and ensures integrity
    # of the data.
    cipher = Cipher(
        algorithms.AES(key_bytes),
        modes.GCM(
            initialized_vector_bytes,
            min_tag_length=ENCRYPTION_AUTH_TAG_LENGTH,
        ),
    )
    encryptor = cipher.encryptor()
    encrypted_value_bytes = encryptor.update(value_bytes) + encryptor.finalize()
    encrypted_password_bytes = public_key.encrypt(
        password_bytes,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )
    return {
        'encrypted_value': base64.b64encode(encrypted_value_bytes + encryptor.tag).decode('utf-8'),
        'encrypted_password': base64.b64encode(encrypted_password_bytes).decode('utf-8'),
    }


def private_decrypt(
    encrypted_password: str,
    encrypted_value: str,
    *,
    private_key: rsa.RSAPrivateKey,
) -> str:
    """Decrypts the given encrypted value using the private key and password.

    Args:
        encrypted_password: Password used to encrypt the private key encoded as base64 string.
        encrypted_value: Encrypted value to decrypt as base64 string.
        private_key: Private key to use for decryption.

    Returns: Decrypted value.
    """
    encrypted_password_bytes = base64.b64decode(encrypted_password.encode('utf-8'))
    encrypted_value_bytes = base64.b64decode(encrypted_value.encode('utf-8'))

    # Decrypt the password
    password_bytes = private_key.decrypt(
        encrypted_password_bytes,
        padding.OAEP(
            mgf=padding.MGF1(algorithm=hashes.SHA1()),
            algorithm=hashes.SHA1(),
            label=None,
        ),
    )

    if len(password_bytes) != ENCRYPTION_KEY_LENGTH + ENCRYPTION_IV_LENGTH:
        raise ValueError('Decryption failed, invalid password length!')

    # Slice the encrypted into cypher and authentication tag
    authentication_tag_bytes = encrypted_value_bytes[-ENCRYPTION_AUTH_TAG_LENGTH:]
    encrypted_data_bytes = encrypted_value_bytes[: (len(encrypted_value_bytes) - ENCRYPTION_AUTH_TAG_LENGTH)]
    encryption_key_bytes = password_bytes[:ENCRYPTION_KEY_LENGTH]
    initialization_vector_bytes = password_bytes[ENCRYPTION_KEY_LENGTH:]

    try:
        cipher = Cipher(
            algorithms.AES(encryption_key_bytes), modes.GCM(initialization_vector_bytes, authentication_tag_bytes)
        )
        decryptor = cipher.decryptor()
        decipher_bytes = decryptor.update(encrypted_data_bytes) + decryptor.finalize()
    except InvalidTagException as exc:
        raise ValueError('Decryption failed, malformed encrypted value or password.') from exc
    except Exception:
        raise

    return decipher_bytes.decode('utf-8')


def load_private_key(private_key_file_base64: str, private_key_password: str) -> rsa.RSAPrivateKey:
    private_key = serialization.load_pem_private_key(
        base64.b64decode(private_key_file_base64.encode('utf-8')),
        password=private_key_password.encode('utf-8'),
    )
    if not isinstance(private_key, rsa.RSAPrivateKey):
        raise TypeError('Invalid private key.')

    return private_key


def _load_public_key(public_key_file_base64: str) -> rsa.RSAPublicKey:
    public_key = serialization.load_pem_public_key(base64.b64decode(public_key_file_base64.encode('utf-8')))
    if not isinstance(public_key, rsa.RSAPublicKey):
        raise TypeError('Invalid public key.')

    return public_key


def decrypt_input_secrets(private_key: rsa.RSAPrivateKey, input_data: Any) -> Any:
    """Decrypt input secrets."""
    if not isinstance(input_data, dict):
        return input_data

    for key, value in input_data.items():
        if isinstance(value, str):
            match = ENCRYPTED_INPUT_VALUE_REGEXP.fullmatch(value)
            if match:
                prefix = match.group(1)
                encrypted_password = match.group(3)
                encrypted_value = match.group(4)
                decrypted_value = private_decrypt(
                    encrypted_password,
                    encrypted_value,
                    private_key=private_key,
                )

                if prefix == ENCRYPTED_STRING_VALUE_PREFIX:
                    input_data[key] = decrypted_value
                elif prefix == ENCRYPTED_JSON_VALUE_PREFIX:
                    input_data[key] = json.loads(decrypted_value)

    return input_data


CHARSET = string.digits + string.ascii_letters


def encode_base62(num: int) -> str:
    """Encode the given number to base62."""
    if num == 0:
        return CHARSET[0]

    res = ''
    while num > 0:
        num, remainder = divmod(num, 62)
        res = CHARSET[remainder] + res
    return res


def create_hmac_signature(secret_key: str, message: str) -> str:
    """Generate an HMAC signature and encodes it using Base62. Base62 encoding reduces the signature length.

    HMAC signature is truncated to 30 characters to make it shorter.

    Args:
        secret_key: Secret key used for signing signatures.
        message: Message to be signed.

    Returns:
        Base62 encoded signature.
    """
    signature = hmac.new(secret_key.encode('utf-8'), message.encode('utf-8'), hashlib.sha256).hexdigest()[:30]

    decimal_signature = int(signature, 16)

    return encode_base62(decimal_signature)



import asyncio

from apify import Actor


async def main() -> None:
    async with Actor:
        # Get input
        actor_input = await Actor.get_input()
        Actor.log.info('Actor input: %s', actor_input)

        # Your Actor logic here
        data = {'message': 'Hello from Actor!', 'input': actor_input}
        await Actor.push_data(data)

        # Set status message
        await Actor.set_status_message('Actor completed successfully')


if __name__ == '__main__':
    asyncio.run(main())



from __future__ import annotations

from urllib.parse import urljoin

from bs4 import BeautifulSoup
from httpx import AsyncClient

from apify import Actor, Request


async def main() -> None:
    # Enter the context of the Actor.
    async with Actor:
        # Retrieve the Actor input, and use default values if not provided.
        actor_input = await Actor.get_input() or {}
        start_urls = actor_input.get('start_urls', [{'url': 'https://apify.com'}])
        max_depth = actor_input.get('max_depth', 1)

        # Exit if no start URLs are provided.
        if not start_urls:
            Actor.log.info('No start URLs specified in Actor input, exiting...')
            await Actor.exit()

        # Open the default request queue for handling URLs to be processed.
        request_queue = await Actor.open_request_queue()

        # Enqueue the start URLs with an initial crawl depth of 0.
        for start_url in start_urls:
            url = start_url.get('url')
            Actor.log.info(f'Enqueuing {url} ...')
            new_request = Request.from_url(url, user_data={'depth': 0})
            await request_queue.add_request(new_request)

        # Create an HTTPX client to fetch the HTML content of the URLs.
        async with AsyncClient() as client:
            # Process the URLs from the request queue.
            while request := await request_queue.fetch_next_request():
                url = request.url

                if not isinstance(request.user_data['depth'], (str, int)):
                    raise TypeError('Request.depth is an unexpected type.')

                depth = int(request.user_data['depth'])
                Actor.log.info(f'Scraping {url} (depth={depth}) ...')

                try:
                    # Fetch the HTTP response from the specified URL using HTTPX.
                    response = await client.get(url, follow_redirects=True)

                    # Parse the HTML content using Beautiful Soup.
                    soup = BeautifulSoup(response.content, 'html.parser')

                    # If the current depth is less than max_depth, find nested links
                    # and enqueue them.
                    if depth < max_depth:
                        for link in soup.find_all('a'):
                            link_href = link.get('href')
                            link_url = urljoin(url, link_href)

                            if link_url.startswith(('http://', 'https://')):
                                Actor.log.info(f'Enqueuing {link_url} ...')
                                new_request = Request.from_url(
                                    link_url,
                                    user_data={'depth': depth + 1},
                                )
                                await request_queue.add_request(new_request)

                    # Extract the desired data.
                    data = {
                        'url': url,
                        'title': soup.title.string if soup.title else None,
                        'h1s': [h1.text for h1 in soup.find_all('h1')],
                        'h2s': [h2.text for h2 in soup.find_all('h2')],
                        'h3s': [h3.text for h3 in soup.find_all('h3')],
                    }

                    # Store the extracted data to the default dataset.
                    await Actor.push_data(data)

                except Exception:
                    Actor.log.exception(f'Cannot extract data from {url}.')

                finally:
                    # Mark the request as handled to ensure it is not processed again.
                    await request_queue.mark_request_as_handled(new_request)



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock

import pytest

from apify_client import ApifyClientAsync
from apify_shared.consts import ApifyEnvVars

from apify import Actor

if TYPE_CHECKING:
    from pytest_httpserver import HTTPServer
    from werkzeug import Request, Response

    from ..conftest import ApifyClientAsyncPatcher

DUMMY_PASSWORD = 'DUMMY_PASSWORD'


@pytest.fixture
def patched_apify_client(apify_client_async_patcher: ApifyClientAsyncPatcher) -> ApifyClientAsync:
    apify_client_async_patcher.patch('user', 'get', return_value={'proxy': {'password': DUMMY_PASSWORD}})
    return ApifyClientAsync()


@pytest.mark.usefixtures('patched_impit_client')
async def test_basic_proxy_configuration_creation(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    patched_apify_client: ApifyClientAsync,
) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    monkeypatch.setenv(ApifyEnvVars.TOKEN.value, 'DUMMY_TOKEN')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    call_mock = Mock()

    def request_handler(request: Request, response: Response) -> Response:
        call_mock(request.url)
        return response

    httpserver.expect_oneshot_request('/').with_post_hook(request_handler).respond_with_json(
        {
            'connected': True,
            'connectionError': None,
            'isManInTheMiddle': True,
        },
        status=200,
    )

    groups = ['GROUP1', 'GROUP2']
    country_code = 'US'

    await Actor.init()

    proxy_configuration = await Actor.create_proxy_configuration(groups=groups, country_code=country_code)

    assert proxy_configuration is not None
    assert proxy_configuration._groups == groups
    assert proxy_configuration._password == DUMMY_PASSWORD
    assert proxy_configuration._country_code == country_code

    assert len(patched_apify_client.calls['user']['get']) == 1  # ty: ignore[unresolved-attribute]
    assert call_mock.call_count == 1

    await Actor.exit()


@pytest.mark.usefixtures('patched_impit_client')
async def test_proxy_configuration_with_actor_proxy_input(
    monkeypatch: pytest.MonkeyPatch,
    httpserver: HTTPServer,
    patched_apify_client: ApifyClientAsync,
) -> None:
    dummy_proxy_status_url = str(httpserver.url_for('/')).removesuffix('/')
    dummy_proxy_url = 'http://dummy-proxy.com:8000'

    monkeypatch.setenv(ApifyEnvVars.TOKEN.value, 'DUMMY_TOKEN')
    monkeypatch.setenv(ApifyEnvVars.PROXY_STATUS_URL.value, dummy_proxy_status_url)

    call_mock = Mock()

    def request_handler(request: Request, response: Response) -> Response:
        call_mock(request.url)
        return response

    httpserver.expect_request('/').with_post_hook(request_handler).respond_with_json(
        {
            'connected': True,
            'connectionError': None,
            'isManInTheMiddle': True,
        },
        status=200,
    )

    await Actor.init()

    proxy_configuration = await Actor.create_proxy_configuration(actor_proxy_input={})
    assert proxy_configuration is None

    proxy_configuration = await Actor.create_proxy_configuration(
        actor_proxy_input={
            'useApifyProxy': False,
        }
    )
    assert proxy_configuration is None

    proxy_configuration = await Actor.create_proxy_configuration(
        actor_proxy_input={
            'proxyUrls': [],
        }
    )
    assert proxy_configuration is None

    proxy_configuration = await Actor.create_proxy_configuration(
        actor_proxy_input={
            'useApifyProxy': False,
            'proxyUrls': [dummy_proxy_url],
        }
    )
    assert proxy_configuration is not None
    assert await proxy_configuration.new_url() == dummy_proxy_url

    proxy_configuration = await Actor.create_proxy_configuration(
        actor_proxy_input={
            'useApifyProxy': True,
        }
    )
    assert proxy_configuration is not None
    assert await proxy_configuration.new_url() == f'http://auto:{DUMMY_PASSWORD}@proxy.apify.com:8000'

    groups = ['GROUP1', 'GROUP2']
    country_code = 'US'
    proxy_configuration = await Actor.create_proxy_configuration(
        actor_proxy_input={
            'useApifyProxy': True,
            'apifyProxyGroups': groups,
            'apifyProxyCountry': country_code,
        }
    )
    assert proxy_configuration is not None
    assert (
        await proxy_configuration.new_url()
        == f'http://groups-{"+".join(groups)},country-{country_code}:{DUMMY_PASSWORD}@proxy.apify.com:8000'
    )

    assert len(patched_apify_client.calls['user']['get']) == 2  # ty: ignore[unresolved-attribute]
    assert call_mock.call_count == 2

    await Actor.exit()



from __future__ import annotations

from typing import TYPE_CHECKING

from apify import Actor

if TYPE_CHECKING:
    from .conftest import MakeActorFunction, RunActorFunction


async def test_create_basic_proxy_configuration(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        groups = ['SHADER']
        country_code = 'US'

        async with Actor:
            proxy_configuration = await Actor.create_proxy_configuration(
                groups=groups,
                country_code=country_code,
            )

            assert proxy_configuration is not None
            assert proxy_configuration._groups == groups
            assert proxy_configuration._password is not None
            assert proxy_configuration._country_code == country_code

    actor = await make_actor(label='proxy-configuration', main_func=main)
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'


async def test_create_proxy_configuration_with_groups_and_country(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        await Actor.init()

        proxy_url_suffix = (
            f'{Actor.configuration.proxy_password}@{Actor.configuration.proxy_hostname}:'
            f'{Actor.configuration.proxy_port}'
        )
        proxy_configuration = await Actor.create_proxy_configuration(actor_proxy_input={'useApifyProxy': True})

        assert proxy_configuration is not None

        new_url = await proxy_configuration.new_url()
        assert new_url == f'http://auto:{proxy_url_suffix}'

        groups = ['SHADER', 'BUYPROXIES94952']
        country_code = 'US'
        proxy_configuration = await Actor.create_proxy_configuration(
            actor_proxy_input={
                'useApifyProxy': True,
                'apifyProxyGroups': groups,
                'apifyProxyCountry': country_code,
            }
        )
        assert proxy_configuration is not None

        new_url = await proxy_configuration.new_url()
        assert new_url == f'http://groups-{"+".join(groups)},country-{country_code}:{proxy_url_suffix}'

        await Actor.exit()

    actor = await make_actor(label='proxy-configuration', main_func=main)
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'



from __future__ import annotations

import asyncio
import builtins
import inspect
import sys
from collections.abc import Callable
from contextlib import asynccontextmanager
from enum import Enum
from functools import wraps
from importlib import metadata
from typing import TYPE_CHECKING, Any, Literal, TypeVar, cast

if TYPE_CHECKING:
    from collections.abc import AsyncIterator

T = TypeVar('T', bound=Callable[..., Any])


def ensure_context(attribute_name: str) -> Callable[[T], T]:
    """Create a decorator that ensures the context manager is initialized before executing the method.

    The decorator checks if the calling instance has the specified attribute and verifies that it is set to `True`.
    If the instance is inactive, it raises a `RuntimeError`. Works for both synchronous and asynchronous methods.

    Args:
        attribute_name: The name of the boolean attribute to check on the instance.

    Returns:
        A decorator that wraps methods with context checking.
    """

    def decorator(method: T) -> T:
        @wraps(method)
        def sync_wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
            if not getattr(self, attribute_name, False):
                raise RuntimeError(f'The {self.__class__.__name__} is not active. Use it within the context.')

            return method(self, *args, **kwargs)

        @wraps(method)
        async def async_wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
            if not getattr(self, attribute_name, False):
                raise RuntimeError(f'The {self.__class__.__name__} is not active. Use it within the async context.')

            return await method(self, *args, **kwargs)

        return cast('T', async_wrapper if inspect.iscoroutinefunction(method) else sync_wrapper)

    return decorator


def get_system_info() -> dict:
    python_version = '.'.join([str(x) for x in sys.version_info[:3]])

    system_info: dict[str, str | bool] = {
        'apify_sdk_version': metadata.version('apify'),
        'apify_client_version': metadata.version('apify-client'),
        'crawlee_version': metadata.version('crawlee'),
        'python_version': python_version,
        'os': sys.platform,
    }

    system_info['is_running_in_ipython'] = is_running_in_ipython()

    return system_info


def is_running_in_ipython() -> bool:
    return getattr(builtins, '__IPYTHON__', False)


# The order of the rendered API groups is defined in the website/docusaurus.config.js file.
GroupName = Literal[
    'Actor',
    'Charging',
    'Configuration',
    'Event data',
    'Event managers',
    'Events',
    'Request loaders',
    'Storage clients',
    'Storage data',
    'Storages',
]


def docs_group(group_name: GroupName) -> Callable:  # noqa: ARG001
    """Mark a symbol for rendering and grouping in documentation.

    This decorator is used solely for documentation purposes and does not modify the behavior
    of the decorated callable.

    Args:
        group_name: The documentation group to which the symbol belongs.

    Returns:
        The original callable without modification.
    """

    def wrapper(func: Callable) -> Callable:
        return func

    return wrapper


def docs_name(symbol_name: str) -> Callable:  # noqa: ARG001
    """Rename a symbol for documentation rendering.

    This decorator modifies only the displayed name of the symbol in the generated documentation
    and does not affect its runtime behavior.

    Args:
        symbol_name: The name to be used in the documentation.

    Returns:
        The original callable without modification.
    """

    def wrapper(func: Callable) -> Callable:
        return func

    return wrapper


def maybe_extract_enum_member_value(maybe_enum_member: Any) -> Any:
    """Extract the value of an enumeration member if it is an Enum, otherwise return the original value."""
    if isinstance(maybe_enum_member, Enum):
        return maybe_enum_member.value
    return maybe_enum_member


class ReentrantLock:
    """A reentrant lock implementation for asyncio using asyncio.Lock."""

    def __init__(self) -> None:
        self._lock = asyncio.Lock()
        self._owner: asyncio.Task | None = None

    @asynccontextmanager
    async def __call__(self) -> AsyncIterator[None]:
        """Acquire the lock if it's not already owned by the current task, otherwise proceed without acquiring."""
        me = asyncio.current_task()
        if me is None:
            raise RuntimeError('ReentrantLock must be used within an asyncio.Task')
        if self._owner is me:
            yield
            return
        async with self._lock:
            self._owner = me
            try:
                yield
            finally:
                self._owner = None



from __future__ import annotations  # noqa: I001

from typing import TYPE_CHECKING, Any

from itemloaders.processors import MapCompose, TakeFirst
from scrapy import Request, Spider
from scrapy.loader import ItemLoader

from src.items import ProductItem  # ty: ignore[unresolved-import]

if TYPE_CHECKING:
    from collections.abc import Generator

    from scrapy.http.response import Response


class ProductItemLoader(ItemLoader):
    default_item_class = ProductItem
    default_output_processor = TakeFirst()
    name_in = MapCompose(str.strip)
    price_in = MapCompose(str.strip)
    description_in = MapCompose(str.strip)


class ItemLoaderSpider(Spider):
    name = 'itemloader_spider'

    def __init__(self, start_urls: list[str], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.start_urls = start_urls

    def start_requests(self) -> Generator[Request, None, None]:
        for url in self.start_urls:
            yield Request(url, callback=self.parse)

    def parse(self, response: Response) -> Generator[Request, None, None]:
        for link in response.css('a[href*="/products/"]::attr(href)').getall():
            yield response.follow(link, callback=self.parse_product)

    def parse_product(self, response: Response) -> Generator[ProductItem, None, None]:
        loader = ProductItemLoader(response=response)  # ty: ignore[invalid-argument-type]
        loader.add_value('url', response.url)
        loader.add_css('name', 'h1::text')
        loader.add_css('price', 'span.price::text')
        loader.add_css('description', 'p.description::text')
        yield loader.load_item()



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING

from apify._charging import DEFAULT_DATASET_ITEM_EVENT, charging_manager_ctx

if TYPE_CHECKING:
    from collections.abc import AsyncIterator


class DatasetClientPpeMixin:
    """A mixin for dataset clients to add support for PPE pricing model and tracking synthetic events."""

    def __init__(self) -> None:
        self.is_default_dataset = False

    def _compute_limit_for_push(self, items_count: int) -> int:
        if self.is_default_dataset and (charging_manager := charging_manager_ctx.get()):
            max_charged_count = charging_manager.calculate_max_event_charge_count_within_limit(
                event_name=DEFAULT_DATASET_ITEM_EVENT
            )
            return min(max_charged_count, items_count) if max_charged_count is not None else items_count
        return items_count

    async def _charge_for_items(self, count_items: int) -> None:
        if self.is_default_dataset and (charging_manager := charging_manager_ctx.get()):
            await charging_manager.charge(
                event_name=DEFAULT_DATASET_ITEM_EVENT,
                count=count_items,
            )

    @asynccontextmanager
    async def _charge_lock(self) -> AsyncIterator[None]:
        """Context manager to acquire the charge lock if PPE charging manager is active."""
        charging_manager = charging_manager_ctx.get()
        if charging_manager:
            async with charging_manager.charge_lock():
                yield
        else:
            yield



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from scrapy import Request, Spider

if TYPE_CHECKING:
    from collections.abc import Generator

    from scrapy.http.response import Response


class BasicSpider(Spider):
    name = 'basic_spider'

    def __init__(self, start_urls: list[str], *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.start_urls = start_urls

    def start_requests(self) -> Generator[Request, None, None]:
        for url in self.start_urls:
            yield Request(url, callback=self.parse)

    def parse(self, response: Response) -> Generator[dict | Request, None, None]:
        for link in response.css('a[href*="/products/"]::attr(href)').getall():
            yield response.follow(link, callback=self.parse_product)

    def parse_product(self, response: Response) -> Generator[dict, None, None]:
        yield {
            'url': response.url,
            'name': response.css('h1::text').get(''),
            'price': response.css('span.price::text').get(''),
            'description': response.css('p.description::text').get(''),
        }



import asyncio

from apify import Actor


async def main() -> None:
    async with Actor:
        # Lifecycle of redirected logs is handled by the context manager.
        async with await Actor.apify_client.run('some_actor_id').get_streamed_log(
            # Redirect all logs from the start of that run, even the logs from past.
            from_start=True
        ):
            await asyncio.sleep(5)
            # Logging will stop out of context

        # Lifecycle of redirected logs can be handled manually.
        streamed_log = await Actor.apify_client.run('some_id').get_streamed_log(
            # Do not redirect historical logs from this actor run.
            # Redirect only new logs from now on.
            from_start=False
        )
        streamed_log.start()
        await asyncio.sleep(5)
        await streamed_log.stop()


if __name__ == '__main__':
    asyncio.run(main())



from apify import Actor


async def main() -> None:
    async with Actor:
        # Open a dataset and write some data in it
        dataset = await Actor.open_dataset(name='my-cool-dataset')
        await dataset.push_data([{'itemNo': i} for i in range(1000)])

        # Read back the first half of the data
        first_half = await dataset.get_data(limit=500)
        Actor.log.info(f'The first half of items = {first_half.items}')

        # Iterate over the second half
        second_half = [item async for item in dataset.iterate_items(offset=500)]
        Actor.log.info(f'The second half of items = {second_half}')



from __future__ import annotations

import os
from typing import TYPE_CHECKING

import pytest

from apify_client import ApifyClientAsync
from apify_shared.consts import ApifyEnvVars
from crawlee import service_locator

import apify._actor
from apify import Actor
from apify.storage_clients import ApifyStorageClient
from apify.storage_clients._apify._alias_resolving import AliasResolver
from apify.storages import RequestQueue

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable
    from pathlib import Path

_TOKEN_ENV_VAR = 'APIFY_TEST_USER_API_TOKEN'
_API_URL_ENV_VAR = 'APIFY_INTEGRATION_TESTS_API_URL'


@pytest.fixture(scope='session')
def apify_token() -> str:
    api_token = os.getenv(_TOKEN_ENV_VAR)

    if not api_token:
        raise RuntimeError(f'{_TOKEN_ENV_VAR} environment variable is missing, cannot run tests!')

    return api_token


@pytest.fixture(scope='session')
def apify_client_async(apify_token: str) -> ApifyClientAsync:
    """Create an instance of the ApifyClientAsync."""
    api_url = os.getenv(_API_URL_ENV_VAR)

    return ApifyClientAsync(apify_token, api_url=api_url)


@pytest.fixture
def prepare_test_env(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> Callable[[], None]:
    """Prepare the testing environment by resetting the global state before each test.

    This fixture ensures that the global state of the package is reset to a known baseline before each test runs.
    It also configures a temporary storage directory for test isolation.

    Args:
        monkeypatch: Test utility provided by pytest for patching.
        tmp_path: A unique temporary directory path provided by pytest for test isolation.

    Returns:
        A callable that prepares the test environment.
    """

    def _prepare_test_env() -> None:
        if hasattr(apify._actor.Actor, '__wrapped__'):
            delattr(apify._actor.Actor, '__wrapped__')

        apify._actor.Actor._active = False

        # Set the environment variable for the local storage directory to the temporary path.
        monkeypatch.setenv(ApifyEnvVars.LOCAL_STORAGE_DIR, str(tmp_path))

        # Reset the services in the service locator.
        service_locator._configuration = None
        service_locator._event_manager = None
        service_locator._storage_client = None
        service_locator.storage_instance_manager.clear_cache()

        # Reset the AliasResolver class state.
        AliasResolver._alias_map = {}
        AliasResolver._alias_init_lock = None

        # Verify that the test environment was set up correctly.
        assert os.environ.get(ApifyEnvVars.LOCAL_STORAGE_DIR) == str(tmp_path)

    return _prepare_test_env


@pytest.fixture(params=['single', 'shared'])
async def request_queue_apify(
    apify_token: str, monkeypatch: pytest.MonkeyPatch, request: pytest.FixtureRequest
) -> AsyncGenerator[RequestQueue]:
    """Create an instance of the Apify request queue on the platform and drop it when the test is finished."""
    monkeypatch.setenv(ApifyEnvVars.TOKEN, apify_token)

    async with Actor:
        rq = await RequestQueue.open(storage_client=ApifyStorageClient(request_queue_access=request.param))
        yield rq
        await rq.drop()


@pytest.fixture(autouse=True)
def _isolate_test_environment(prepare_test_env: Callable[[], None]) -> None:
    """Isolate the testing environment by resetting global state before each test.

    This fixture ensures that each test starts with a clean slate and that any modifications during the test
    do not affect subsequent tests. It runs automatically for all tests.

    Args:
        prepare_test_env: Fixture to prepare the environment before each test.
    """
    prepare_test_env()



from __future__ import annotations

import codecs
import pickle
from logging import getLogger
from typing import Any, cast

from scrapy import Request as ScrapyRequest
from scrapy import Spider
from scrapy.http.headers import Headers
from scrapy.utils.request import request_from_dict

from crawlee._request import UserData
from crawlee._types import HttpHeaders

from apify import Request as ApifyRequest

logger = getLogger(__name__)


def to_apify_request(scrapy_request: ScrapyRequest, spider: Spider) -> ApifyRequest | None:
    """Convert a Scrapy request to an Apify request.

    Args:
        scrapy_request: The Scrapy request to be converted.
        spider: The Scrapy spider that the request is associated with.

    Returns:
        The converted Apify request if the conversion was successful, otherwise None.
    """
    if not isinstance(scrapy_request, ScrapyRequest):
        logger.warning('Failed to convert to Apify request: Scrapy request must be a ScrapyRequest instance.')
        return None

    logger.debug(f'to_apify_request was called (scrapy_request={scrapy_request})...')

    # Configuration to behave as similarly as possible to Scrapy's default RFPDupeFilter.
    request_kwargs: dict[str, Any] = {
        'url': scrapy_request.url,
        'method': scrapy_request.method,
        'payload': scrapy_request.body,
        'use_extended_unique_key': True,
        'keep_url_fragment': False,
    }

    try:
        if scrapy_request.dont_filter:
            request_kwargs['always_enqueue'] = True
        else:
            if scrapy_request.meta.get('apify_request_unique_key'):
                request_kwargs['unique_key'] = scrapy_request.meta['apify_request_unique_key']

            if scrapy_request.meta.get('apify_request_id'):
                request_kwargs['id'] = scrapy_request.meta['apify_request_id']

        user_data = scrapy_request.meta.get('userData', {})

        # Convert UserData Pydantic model to a plain dict to prevent CrawleeRequestData objects
        # from leaking into Request.from_url() during Scrapy-Apify roundtrips.
        if isinstance(user_data, UserData):
            user_data = user_data.model_dump(by_alias=True)

        # Remove internal Crawlee data since it's managed by Request.from_url() and values
        # from previous roundtrips cause incorrect state.
        if isinstance(user_data, dict):
            user_data.pop('__crawlee', None)

        request_kwargs['user_data'] = user_data if isinstance(user_data, dict) else {}

        # Convert Scrapy's headers to a HttpHeaders and store them in the apify_request
        if isinstance(scrapy_request.headers, Headers):
            request_kwargs['headers'] = HttpHeaders(dict(scrapy_request.headers.to_unicode_dict()))
        else:
            logger.warning(
                f'Invalid scrapy_request.headers type, not scrapy.http.headers.Headers: {scrapy_request.headers}'
            )

        apify_request = ApifyRequest.from_url(**request_kwargs)

        # Serialize the Scrapy ScrapyRequest and store it in the apify_request.
        #   - This process involves converting the Scrapy ScrapyRequest object into a dictionary, encoding it to base64,
        #     and storing it as 'scrapy_request' within the 'userData' dictionary of the apify_request.
        #   - The serialization process can be referenced at: https://stackoverflow.com/questions/30469575/.
        scrapy_request_dict = scrapy_request.to_dict(spider=spider)
        scrapy_request_dict_encoded = codecs.encode(pickle.dumps(scrapy_request_dict), 'base64').decode()
        apify_request.user_data['scrapy_request'] = scrapy_request_dict_encoded

    except Exception as exc:
        logger.warning(f'Conversion of Scrapy request {scrapy_request} to Apify request failed; {exc}')
        return None

    logger.debug(f'scrapy_request was converted to the apify_request={apify_request}')
    return apify_request


def to_scrapy_request(apify_request: ApifyRequest, spider: Spider) -> ScrapyRequest:
    """Convert an Apify request to a Scrapy request.

    Args:
        apify_request: The Apify request to be converted.
        spider: The Scrapy spider that the request is associated with.

    Raises:
        TypeError: If the Apify request is not an instance of the `ApifyRequest` class.
        ValueError: If the Apify request does not contain the required keys.

    Returns:
        The converted Scrapy request.
    """
    if not isinstance(cast('Any', apify_request), ApifyRequest):
        raise TypeError('apify_request must be a crawlee.ScrapyRequest instance')

    logger.debug(f'to_scrapy_request was called (apify_request={apify_request})...')

    # If the apify_request comes from the Scrapy
    if 'scrapy_request' in apify_request.user_data:
        # Deserialize the Scrapy ScrapyRequest from the apify_request.
        #   - This process involves decoding the base64-encoded request data and reconstructing
        #     the Scrapy ScrapyRequest object from its dictionary representation.
        logger.debug('Restoring the Scrapy ScrapyRequest from the apify_request...')

        scrapy_request_dict_encoded = apify_request.user_data['scrapy_request']
        if not isinstance(scrapy_request_dict_encoded, str):
            raise TypeError('scrapy_request_dict_encoded must be a string')

        scrapy_request_dict = pickle.loads(codecs.decode(scrapy_request_dict_encoded.encode(), 'base64'))
        if not isinstance(scrapy_request_dict, dict):
            raise TypeError('scrapy_request_dict must be a dictionary')

        scrapy_request = request_from_dict(scrapy_request_dict, spider=spider)
        if not isinstance(scrapy_request, ScrapyRequest):
            raise TypeError('scrapy_request must be an instance of the ScrapyRequest class')

        logger.debug(f'Scrapy ScrapyRequest successfully reconstructed (scrapy_request={scrapy_request})...')

        # Update the meta field with the meta field from the apify_request
        meta = scrapy_request.meta or {}
        meta.update({'apify_request_unique_key': apify_request.unique_key})
        # scrapy_request.meta is a property, so we have to set it like this
        scrapy_request._meta = meta  # noqa: SLF001

    # If the apify_request comes directly from the Scrapy, typically start URLs.
    else:
        logger.debug('Gonna create a new Scrapy ScrapyRequest (cannot be restored)')

        scrapy_request = ScrapyRequest(
            url=apify_request.url,
            method=apify_request.method,
            meta={
                'apify_request_unique_key': apify_request.unique_key,
            },
        )

    # Add optional 'headers' field
    if apify_request.headers:
        scrapy_request.headers |= Headers(apify_request.headers)

    # Add optional 'userData' field
    if apify_request.user_data:
        scrapy_request.meta['userData'] = apify_request.user_data

    logger.debug(f'an apify_request was converted to the scrapy_request={scrapy_request}')
    return scrapy_request



from __future__ import annotations

import os

from .main import main
from apify.scrapy import initialize_logging, run_scrapy_actor

os.environ['SCRAPY_SETTINGS_MODULE'] = 'src.settings'

if __name__ == '__main__':
    initialize_logging()
    run_scrapy_actor(main())



from __future__ import annotations

import asyncio
import contextlib
import logging
from typing import TYPE_CHECKING
from unittest.mock import AsyncMock

import pytest

from apify_shared.consts import ActorExitCodes, ApifyEnvVars
from crawlee.events._types import Event

from apify import Actor

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable

    from apify._actor import _ActorType


@pytest.fixture(
    params=[
        pytest.param(('instance', 'manual'), id='instance-manual'),
        pytest.param(('instance', 'async_with'), id='instance-async-with'),
        pytest.param(('class', 'manual'), id='class-manual'),
        pytest.param(('class', 'async_with'), id='class-async-with'),
    ]
)
async def actor(
    request: pytest.FixtureRequest,
) -> AsyncGenerator[_ActorType, None]:
    """Yield Actor instance or class in different initialization modes.

    - instance-manual: Actor() with manual init()/exit()
    - instance-async-with: Actor() used as async context manager
    - class-manual: Actor class with manual init()/exit()
    - class-async-with: Actor class used as async context manager

    Each Actor is properly initialized before yielding and cleaned up after.
    """
    scope, mode = request.param

    if scope == 'instance':
        if mode == 'manual':
            instance = Actor()
            await instance.init()
            yield instance
            await instance.exit()
        else:
            async with Actor() as instance:
                yield instance

    elif scope == 'class':
        if mode == 'manual':
            await Actor.init()
            yield Actor
            await Actor.exit()
        else:
            async with Actor:
                yield Actor

    else:
        raise ValueError(f'Unknown scope: {scope}')


async def test_actor_init_instance_manual() -> None:
    """Test that Actor instance can be properly initialized and cleaned up manually."""
    actor = Actor()
    await actor.init()
    assert actor._active is True
    await actor.exit()
    assert actor._active is False


async def test_actor_init_instance_async_with() -> None:
    """Test that Actor instance can be properly initialized and cleaned up using async context manager."""
    actor = Actor()
    async with actor:
        assert actor._active is True

    assert actor._active is False


async def test_actor_init_class_manual() -> None:
    """Test that Actor class can be properly initialized and cleaned up manually."""
    await Actor.init()
    assert Actor._active is True
    await Actor.exit()
    assert not Actor._active


async def test_actor_init_class_async_with() -> None:
    """Test that Actor class can be properly initialized and cleaned up using async context manager."""
    async with Actor:
        assert Actor._active is True

    assert not Actor._active


async def test_fail_properly_deinitializes_actor(actor: _ActorType) -> None:
    """Test that fail() method properly deinitializes the Actor."""
    assert actor._active
    await actor.fail()
    assert actor._active is False


async def test_actor_handles_exceptions_and_cleans_up_properly() -> None:
    """Test that Actor properly cleans up when an exception occurs in the async context manager."""
    actor = None

    with contextlib.suppress(Exception):
        async with Actor() as actor:
            assert actor._active
            raise Exception('Failed')  # noqa: TRY002

    assert actor is not None
    assert actor._active is False


async def test_double_init_raises_runtime_error(actor: _ActorType) -> None:
    """Test that attempting to initialize an already initialized Actor raises RuntimeError."""
    assert actor._active
    with pytest.raises(RuntimeError):
        await actor.init()


async def test_exit_without_init_raises_runtime_error() -> None:
    """Test that calling exit() on an uninitialized Actor raises RuntimeError."""
    with pytest.raises(RuntimeError):
        await Actor.exit()

    with pytest.raises(RuntimeError):
        await Actor().exit()


async def test_fail_without_init_raises_runtime_error() -> None:
    """Test that calling fail() on an uninitialized Actor raises RuntimeError."""
    with pytest.raises(RuntimeError):
        await Actor.fail()

    with pytest.raises(RuntimeError):
        await Actor().fail()


async def test_reboot_in_local_environment_logs_error_message(
    actor: _ActorType,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test that reboot() logs an error when not running on the Apify platform."""
    await actor.reboot()

    # Check that the error message was logged
    assert 'Actor.reboot() is only supported when running on the Apify platform.' in caplog.text


async def test_exit_sets_custom_exit_code_and_status_message(actor: _ActorType) -> None:
    """Test that exit() properly sets custom exit code and status message."""
    await actor.exit(exit_code=42, status_message='Exiting with code 42')
    assert actor.exit_code == 42
    assert actor.status_message == 'Exiting with code 42'


async def test_fail_sets_custom_exit_code_and_status_message(actor: _ActorType) -> None:
    """Test that fail() properly sets custom exit code and status message."""
    await actor.fail(exit_code=99, status_message='Failing with code 99')
    assert actor.exit_code == 99
    assert actor.status_message == 'Failing with code 99'


async def test_unhandled_exception_sets_error_exit_code() -> None:
    """Test that unhandled exceptions in context manager set the error exit code."""
    actor = Actor(exit_process=False)
    with pytest.raises(RuntimeError):
        async with actor:
            raise RuntimeError('Test error')

    assert actor.exit_code == ActorExitCodes.ERROR_USER_FUNCTION_THREW.value


async def test_actor_stops_periodic_events_after_exit(monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that periodic events (PERSIST_STATE and SYSTEM_INFO) stop emitting after Actor exits."""
    monkeypatch.setenv(ApifyEnvVars.SYSTEM_INFO_INTERVAL_MILLIS, '100')
    monkeypatch.setenv(ApifyEnvVars.PERSIST_STATE_INTERVAL_MILLIS, '100')
    on_persist = []
    on_system_info = []

    def on_event(event_type: Event) -> Callable:
        nonlocal on_persist
        nonlocal on_system_info
        if event_type == Event.PERSIST_STATE:
            return on_persist.append
        if event_type == Event.SYSTEM_INFO:
            return on_system_info.append
        return print

    actor = Actor()
    async with actor:
        assert actor._active
        actor.on(Event.PERSIST_STATE, on_event(Event.PERSIST_STATE))
        actor.on(Event.SYSTEM_INFO, on_event(Event.SYSTEM_INFO))
        await asyncio.sleep(1)

    on_persist_count = len(on_persist)
    on_system_info_count = len(on_system_info)
    assert on_persist_count != 0
    assert on_system_info_count != 0

    # Check if events stopped emitting.
    await asyncio.sleep(0.2)
    assert on_persist_count == len(on_persist)
    assert on_system_info_count == len(on_system_info)


async def test_actor_fail_prevents_further_execution(caplog: pytest.LogCaptureFixture) -> None:
    """Test that calling Actor.fail() prevents further code execution in the Actor context."""
    caplog.set_level(logging.INFO)
    try:
        # Explicitly set exit_process=True since in Pytest env it defaults to False.
        async with Actor(exit_process=True) as actor:
            await actor.fail(exit_code=2, exception=Exception('abc'), status_message='cde')
            raise RuntimeError('This should not trigger')
    except SystemExit:
        fail_records = [r for r in caplog.records if r.msg == 'Actor failed with an exception']
        assert len(fail_records) == 1
        assert fail_records[0].levelno == logging.ERROR
        assert fail_records[0].exc_text == 'Exception: abc'

        exit_records = [r for r in caplog.records if r.msg == 'Exiting Actor']
        assert len(exit_records) == 1
        assert exit_records[0].levelno == logging.INFO

        status_records = [r for r in caplog.records if r.msg == '[Terminal status message]: cde']
        assert len(status_records) == 1
        assert status_records[0].levelno == logging.INFO


@pytest.mark.parametrize(
    ('first_with_call', 'second_with_call'),
    [
        pytest.param(False, False, id='both_without_call'),
        pytest.param(False, True, id='first_without_call'),
        pytest.param(True, False, id='second_without_call'),
        pytest.param(True, True, id='both_with_call'),
    ],
)
async def test_actor_sequential_contexts(*, first_with_call: bool, second_with_call: bool) -> None:
    """Test that Actor and Actor() can be used in two sequential async context manager blocks."""
    mock = AsyncMock()
    async with Actor(exit_process=False) if first_with_call else Actor as actor:
        await mock()
        assert actor._active is True

    # After exiting the context, new Actor instance can be created without conflicts.
    async with Actor() if second_with_call else Actor as actor:
        await mock()
        assert actor._active is True

    # The mock should have been called twice, once in each context.
    assert mock.call_count == 2



from __future__ import annotations

from scrapy.utils.reactor import install_reactor

# Install Twisted's asyncio reactor before importing any other Twisted or
# Scrapy components.
install_reactor('twisted.internet.asyncioreactor.AsyncioSelectorReactor')

import os

from apify.scrapy import initialize_logging, run_scrapy_actor

# Import your main Actor coroutine here.
from .main import main

# Ensure the location to the Scrapy settings module is defined.
os.environ['SCRAPY_SETTINGS_MODULE'] = 'src.settings'


if __name__ == '__main__':
    initialize_logging()
    run_scrapy_actor(main())



from __future__ import annotations

import asyncio
import json
from typing import TYPE_CHECKING

from apify_shared.consts import ActorPermissionLevel
from crawlee._utils.crypto import crypto_random_object_id

from ._utils import generate_unique_resource_name
from apify import Actor
from apify._models import ActorRun

if TYPE_CHECKING:
    from apify_client import ApifyClientAsync

    from .conftest import MakeActorFunction, RunActorFunction


async def test_actor_reports_running_on_platform(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        async with Actor:
            assert Actor.is_at_home() is True

    actor = await make_actor(label='is-at-home', main_func=main)
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'


async def test_actor_retrieves_env_vars(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        async with Actor:
            env_dict = Actor.get_env()

            assert env_dict.get('is_at_home') is True
            assert env_dict.get('token') is not None
            assert env_dict.get('events_websocket_url') is not None
            assert env_dict.get('input_key') is not None

            assert len(env_dict.get('id', '')) == 17
            assert len(env_dict.get('build_id', '')) == 17
            assert len(env_dict.get('run_id', '')) == 17
            assert len(env_dict.get('user_id', '')) == 17

    actor = await make_actor(label='get-env', main_func=main)
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'


async def test_actor_creates_new_client_instance(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        import os

        from apify_shared.consts import ActorEnvVars

        async with Actor:
            new_client = Actor.new_client()
            assert new_client is not Actor.apify_client

            default_key_value_store_id = os.getenv(ActorEnvVars.DEFAULT_KEY_VALUE_STORE_ID)
            assert default_key_value_store_id is not None
            kv_store_client = new_client.key_value_store(default_key_value_store_id)
            await kv_store_client.set_record('OUTPUT', 'TESTING-OUTPUT')

    actor = await make_actor(label='new-client', main_func=main)
    run_result = await run_actor(actor)

    assert run_result.status == 'SUCCEEDED'

    output_record = await actor.last_run().key_value_store().get_record('OUTPUT')
    assert output_record is not None
    assert output_record['value'] == 'TESTING-OUTPUT'


async def test_actor_sets_status_message(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        async with Actor:
            actor_input = await Actor.get_input() or {}
            await Actor.set_status_message('testing-status-message', **actor_input)

    actor = await make_actor(label='set-status-message', main_func=main)
    run_result_1 = await run_actor(actor)

    assert run_result_1.status == 'SUCCEEDED'
    assert run_result_1.status_message == 'testing-status-message'
    assert run_result_1.is_status_message_terminal is None

    run_result_2 = await run_actor(actor, run_input={'is_terminal': True})

    assert run_result_2.status == 'SUCCEEDED'
    assert run_result_2.status_message == 'testing-status-message'
    assert run_result_2.is_status_message_terminal is True


async def test_actor_starts_another_actor_instance(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main_inner() -> None:
        async with Actor:
            await asyncio.sleep(5)
            actor_input = await Actor.get_input() or {}
            test_value = actor_input.get('test_value')
            await Actor.set_value('OUTPUT', f'{test_value}_XXX_{test_value}')

    async def main_outer() -> None:
        async with Actor:
            actor_input = await Actor.get_input() or {}
            inner_actor_id = actor_input.get('inner_actor_id')
            test_value = actor_input.get('test_value')

            assert inner_actor_id is not None

            await Actor.start(inner_actor_id, run_input={'test_value': test_value})

            inner_run_status = await Actor.apify_client.actor(inner_actor_id).last_run().get()
            assert inner_run_status is not None
            assert inner_run_status.get('status') in ['READY', 'RUNNING']

    inner_actor = await make_actor(label='start-inner', main_func=main_inner)
    outer_actor = await make_actor(label='start-outer', main_func=main_outer)

    inner_actor_id = (await inner_actor.get() or {})['id']
    test_value = crypto_random_object_id()

    run_result_outer = await run_actor(
        outer_actor,
        run_input={'test_value': test_value, 'inner_actor_id': inner_actor_id},
    )

    assert run_result_outer.status == 'SUCCEEDED'

    await inner_actor.last_run().wait_for_finish(wait_secs=600)

    inner_output_record = await inner_actor.last_run().key_value_store().get_record('OUTPUT')
    assert inner_output_record is not None
    assert inner_output_record['value'] == f'{test_value}_XXX_{test_value}'


async def test_actor_calls_another_actor(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main_inner() -> None:
        async with Actor:
            await asyncio.sleep(5)
            actor_input = await Actor.get_input() or {}
            test_value = actor_input.get('test_value')
            await Actor.set_value('OUTPUT', f'{test_value}_XXX_{test_value}')

    async def main_outer() -> None:
        async with Actor:
            actor_input = await Actor.get_input() or {}
            inner_actor_id = actor_input.get('inner_actor_id')
            test_value = actor_input.get('test_value')

            assert inner_actor_id is not None

            await Actor.call(inner_actor_id, run_input={'test_value': test_value})

            inner_run_status = await Actor.apify_client.actor(inner_actor_id).last_run().get()
            assert inner_run_status is not None
            assert inner_run_status.get('status') == 'SUCCEEDED'

    inner_actor = await make_actor(label='call-inner', main_func=main_inner)
    outer_actor = await make_actor(label='call-outer', main_func=main_outer)

    inner_actor_id = (await inner_actor.get() or {})['id']
    test_value = crypto_random_object_id()

    run_result_outer = await run_actor(
        outer_actor,
        run_input={'test_value': test_value, 'inner_actor_id': inner_actor_id},
    )

    assert run_result_outer.status == 'SUCCEEDED'

    await inner_actor.last_run().wait_for_finish(wait_secs=600)

    inner_output_record = await inner_actor.last_run().key_value_store().get_record('OUTPUT')
    assert inner_output_record is not None
    assert inner_output_record['value'] == f'{test_value}_XXX_{test_value}'


async def test_actor_calls_task(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
    apify_client_async: ApifyClientAsync,
) -> None:
    async def main_inner() -> None:
        async with Actor:
            await asyncio.sleep(5)
            actor_input = await Actor.get_input() or {}
            test_value = actor_input.get('test_value')
            await Actor.set_value('OUTPUT', f'{test_value}_XXX_{test_value}')

    async def main_outer() -> None:
        async with Actor:
            actor_input = await Actor.get_input() or {}
            inner_task_id = actor_input.get('inner_task_id')

            assert inner_task_id is not None

            await Actor.call_task(inner_task_id)

            inner_run_status = await Actor.apify_client.task(inner_task_id).last_run().get()
            assert inner_run_status is not None
            assert inner_run_status.get('status') == 'SUCCEEDED'

    inner_actor = await make_actor(label='call-task-inner', main_func=main_inner)
    outer_actor = await make_actor(label='call-task-outer', main_func=main_outer)

    inner_actor_id = (await inner_actor.get() or {})['id']
    test_value = crypto_random_object_id()

    task = await apify_client_async.tasks().create(
        actor_id=inner_actor_id,
        name=generate_unique_resource_name('actor-call-task'),
        task_input={'test_value': test_value},
    )

    run_result_outer = await run_actor(
        outer_actor,
        run_input={'test_value': test_value, 'inner_task_id': task['id']},
        force_permission_level=ActorPermissionLevel.FULL_PERMISSIONS,
    )

    assert run_result_outer.status == 'SUCCEEDED'

    await inner_actor.last_run().wait_for_finish(wait_secs=600)

    inner_output_record = await inner_actor.last_run().key_value_store().get_record('OUTPUT')
    assert inner_output_record is not None
    assert inner_output_record['value'] == f'{test_value}_XXX_{test_value}'

    await apify_client_async.task(task['id']).delete()


async def test_actor_aborts_another_actor_run(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main_inner() -> None:
        async with Actor:
            await asyncio.sleep(180)
            # This should not be set, the actor should be aborted by now
            await Actor.set_value('OUTPUT', 'dummy')

    async def main_outer() -> None:
        async with Actor:
            actor_input = await Actor.get_input() or {}
            inner_run_id = actor_input.get('inner_run_id')

            assert inner_run_id is not None

            await Actor.abort(inner_run_id)

    inner_actor = await make_actor(label='abort-inner', main_func=main_inner)
    outer_actor = await make_actor(label='abort-outer', main_func=main_outer)

    run_result_inner = await inner_actor.start(force_permission_level=ActorPermissionLevel.FULL_PERMISSIONS)
    inner_run_id = run_result_inner['id']

    run_result_outer = await run_actor(
        outer_actor,
        run_input={'inner_run_id': inner_run_id},
        force_permission_level=ActorPermissionLevel.FULL_PERMISSIONS,
    )

    assert run_result_outer.status == 'SUCCEEDED'

    await inner_actor.last_run().wait_for_finish(wait_secs=600)
    inner_actor_last_run_dict = await inner_actor.last_run().get()
    inner_actor_last_run = ActorRun.model_validate(inner_actor_last_run_dict)

    assert inner_actor_last_run.status == 'ABORTED'

    inner_output_record = await inner_actor.last_run().key_value_store().get_record('OUTPUT')
    assert inner_output_record is None


async def test_actor_metamorphs_into_another_actor(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main_inner() -> None:
        import os

        from apify_shared.consts import ActorEnvVars

        async with Actor:
            assert os.getenv(ActorEnvVars.INPUT_KEY) is not None
            assert os.getenv(ActorEnvVars.INPUT_KEY) != 'INPUT'
            actor_input = await Actor.get_input() or {}

            test_value = actor_input.get('test_value', '')
            assert test_value.endswith('_BEFORE_METAMORPH')

            output = test_value.replace('_BEFORE_METAMORPH', '_AFTER_METAMORPH')
            await Actor.set_value('OUTPUT', output)

    async def main_outer() -> None:
        async with Actor:
            actor_input = await Actor.get_input() or {}
            inner_actor_id = actor_input.get('inner_actor_id')
            test_value = actor_input.get('test_value')
            new_test_value = f'{test_value}_BEFORE_METAMORPH'

            assert inner_actor_id is not None

            await Actor.metamorph(inner_actor_id, run_input={'test_value': new_test_value})

            # This should not be called
            await Actor.set_value('RECORD_AFTER_METAMORPH_CALL', 'dummy')
            raise AssertionError('The Actor should have been metamorphed by now')

    inner_actor = await make_actor(label='metamorph-inner', main_func=main_inner)
    outer_actor = await make_actor(label='metamorph-outer', main_func=main_outer)

    inner_actor_id = (await inner_actor.get() or {})['id']
    test_value = crypto_random_object_id()

    run_result_outer = await run_actor(
        outer_actor,
        run_input={'test_value': test_value, 'inner_actor_id': inner_actor_id},
    )

    assert run_result_outer.status == 'SUCCEEDED'

    outer_run_key_value_store = outer_actor.last_run().key_value_store()

    outer_output_record = await outer_run_key_value_store.get_record('OUTPUT')
    assert outer_output_record is not None
    assert outer_output_record['value'] == f'{test_value}_AFTER_METAMORPH'

    assert await outer_run_key_value_store.get_record('RECORD_AFTER_METAMORPH_CALL') is None

    # After metamorph, the run still belongs to the original Actor, so the inner one should have no runs
    assert await inner_actor.last_run().get() is None


async def test_actor_reboots_successfully(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main() -> None:
        async with Actor:
            print('Starting...')
            cnt = await Actor.get_value('reboot_counter', 0)

            if cnt < 2:
                print(f'Rebooting (cnt = {cnt})...')
                await Actor.set_value('reboot_counter', cnt + 1)
                await Actor.reboot()
                await Actor.set_value('THIS_KEY_SHOULD_NOT_BE_WRITTEN', 'XXX')

            print('Finishing...')

    actor = await make_actor(label='actor_rebooter', main_func=main)

    run_result = await run_actor(
        actor,
        run_input={'counter_key': 'reboot_counter'},
    )

    assert run_result.status == 'SUCCEEDED'

    not_written_value = await actor.last_run().key_value_store().get_record('THIS_KEY_SHOULD_NOT_BE_WRITTEN')
    assert not_written_value is None

    reboot_counter = await actor.last_run().key_value_store().get_record('reboot_counter')
    assert reboot_counter is not None
    assert reboot_counter['value'] == 2


async def test_actor_adds_webhook_and_receives_event(
    make_actor: MakeActorFunction,
    run_actor: RunActorFunction,
) -> None:
    async def main_server() -> None:
        import os
        from http.server import BaseHTTPRequestHandler, HTTPServer

        from apify_shared.consts import ActorEnvVars

        webhook_body = ''

        async with Actor:

            class WebhookHandler(BaseHTTPRequestHandler):
                def do_GET(self) -> None:
                    self.send_response(200)
                    self.end_headers()
                    self.wfile.write(bytes('Hello, world!', encoding='utf-8'))

                def do_POST(self) -> None:
                    nonlocal webhook_body
                    content_length = self.headers.get('content-length')
                    length = int(content_length) if content_length else 0
                    webhook_body = self.rfile.read(length).decode('utf-8')
                    self.send_response(200)
                    self.end_headers()
                    self.wfile.write(bytes('Hello, world!', encoding='utf-8'))

            container_port = int(os.getenv(ActorEnvVars.WEB_SERVER_PORT, ''))
            with HTTPServer(('', container_port), WebhookHandler) as server:
                await Actor.set_value('INITIALIZED', value=True)
                while not webhook_body:
                    server.handle_request()

            await Actor.set_value('WEBHOOK_BODY', webhook_body)

    async def main_client() -> None:
        from apify import Webhook, WebhookEventType

        async with Actor:
            actor_input = await Actor.get_input() or {}
            server_actor_container_url = str(actor_input.get('server_actor_container_url'))

            await Actor.add_webhook(
                Webhook(
                    event_types=[WebhookEventType.ACTOR_RUN_SUCCEEDED],
                    request_url=server_actor_container_url,
                )
            )

    server_actor, client_actor = await asyncio.gather(
        make_actor(label='add-webhook-server', main_func=main_server),
        make_actor(label='add-webhook-client', main_func=main_client),
    )

    server_actor_run = await server_actor.start()
    server_actor_container_url = server_actor_run['containerUrl']

    server_actor_initialized = await server_actor.last_run().key_value_store().get_record('INITIALIZED')
    while not server_actor_initialized:
        server_actor_initialized = await server_actor.last_run().key_value_store().get_record('INITIALIZED')
        await asyncio.sleep(1)

    ac_run_result = await run_actor(
        client_actor,
        run_input={'server_actor_container_url': server_actor_container_url},
    )

    assert ac_run_result.status == 'SUCCEEDED'

    sa_run_result_dict = await server_actor.last_run().wait_for_finish(wait_secs=600)
    sa_run_result = ActorRun.model_validate(sa_run_result_dict)

    assert sa_run_result.status == 'SUCCEEDED'

    webhook_body_record = await server_actor.last_run().key_value_store().get_record('WEBHOOK_BODY')
    assert webhook_body_record is not None
    assert webhook_body_record['value'] != ''
    parsed_webhook_body = json.loads(webhook_body_record['value'])

    assert parsed_webhook_body['eventData']['actorId'] == ac_run_result.act_id
    assert parsed_webhook_body['eventData']['actorRunId'] == ac_run_result.id



from __future__ import annotations

from dataclasses import dataclass

import pytest
from scrapy import Field, Item

from apify import Actor
from apify.scrapy.pipelines import ActorDatasetPushPipeline


class DummyItem(Item):
    a = Field()
    b = Field()
    c = Field()


class TitleItem(Item):
    url = Field()
    title = Field()


@pytest.fixture
def pipeline() -> ActorDatasetPushPipeline:
    """Fixture to create an Actor dataset push pipeline."""
    return ActorDatasetPushPipeline()


@dataclass(frozen=True)
class ItemTestCase:
    item: Item
    item_dict: dict
    expected_exception: type[Exception] | None


@pytest.mark.parametrize(
    'tc',
    [
        ItemTestCase(
            item=DummyItem(a='string', b=123, c=False),
            item_dict={'a': 'string', 'b': 123, 'c': False},
            expected_exception=None,
        ),
        ItemTestCase(
            item=TitleItem(url='https://example.com', title='Example'),
            item_dict={'url': 'https://example.com', 'title': 'Example'},
            expected_exception=None,
        ),
        ItemTestCase(
            item=None,  # ty: ignore[invalid-argument-type]
            item_dict={},
            expected_exception=TypeError,
        ),
    ],
    ids=['dummy_item_with_valid_data', 'title_item_with_valid_data', 'none_item_raises_type_error'],
)
async def test_process_item(
    monkeypatch: pytest.MonkeyPatch,
    pipeline: ActorDatasetPushPipeline,
    tc: ItemTestCase,
) -> None:
    dataset = []

    async def mock_push_data(item: dict) -> None:
        dataset.append(item)

    monkeypatch.setattr(Actor, 'push_data', mock_push_data)

    if tc.expected_exception:
        with pytest.raises(tc.expected_exception):
            await pipeline.process_item(tc.item)

    else:
        output = await pipeline.process_item(tc.item)
        assert output == tc.item
        assert dataset == [tc.item_dict]



from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from apify_shared.consts import ApifyEnvVars

from ._utils import generate_unique_resource_name
from apify import Actor
from apify.storage_clients import ApifyStorageClient
from apify.storages import Dataset

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator

    from apify_client import ApifyClientAsync


@pytest.fixture
async def dataset_apify(apify_token: str, monkeypatch: pytest.MonkeyPatch) -> AsyncGenerator[Dataset]:
    """Create an instance of the Apify dataset on the platform and drop it when the test is finished."""
    monkeypatch.setenv(ApifyEnvVars.TOKEN, apify_token)

    async with Actor:
        ds = await Dataset.open(storage_client=ApifyStorageClient())
        yield ds
        await ds.drop()


async def test_push_and_verify_data_in_default_dataset(dataset_apify: Dataset) -> None:
    """Test pushing 100 items and verifying them via API."""
    desired_item_count = 100
    ds = dataset_apify

    await ds.push_data([{'id': i} for i in range(desired_item_count)])

    items = await ds.list_items()
    assert len(items) == desired_item_count
    assert items[0]['id'] == 0
    assert items[-1]['id'] == desired_item_count - 1


async def test_push_large_data_chunks_over_9mb(dataset_apify: Dataset) -> None:
    """Test that SDK chunking logic works for large payloads."""
    ds = dataset_apify

    await ds.push_data([{'str': 'x' * 10000} for _ in range(5000)])  # ~50MB

    items = await ds.list_items()
    assert len(items) == 5000
    for item in items:
        assert item['str'] == 'x' * 10000


async def test_dataset_iter_items(dataset_apify: Dataset) -> None:
    """Test iteration over dataset items."""
    inserted_data = {'Something': 'something else'}
    ds = dataset_apify

    await ds.push_data(inserted_data)
    requested_data = [item async for item in ds.iterate_items()]

    assert len(requested_data) == 1
    assert requested_data[0] == inserted_data


async def test_same_references_in_default_dataset(apify_token: str, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that opening the default dataset twice returns the same instance."""
    monkeypatch.setenv(ApifyEnvVars.TOKEN, apify_token)

    async with Actor:
        dataset1 = await Actor.open_dataset()
        dataset2 = await Actor.open_dataset()
        assert dataset1 is dataset2


async def test_same_references_in_named_dataset(apify_token: str, monkeypatch: pytest.MonkeyPatch) -> None:
    """Test that opening a named dataset by name and then by ID returns the same instance."""
    dataset_name = generate_unique_resource_name('dataset')
    monkeypatch.setenv(ApifyEnvVars.TOKEN, apify_token)

    async with Actor:
        dataset_by_name_1 = await Actor.open_dataset(name=dataset_name)
        try:
            dataset_by_name_2 = await Actor.open_dataset(name=dataset_name)
            assert dataset_by_name_1 is dataset_by_name_2

            dataset_1_metadata = await dataset_by_name_1.get_metadata()
            dataset_by_id_1 = await Actor.open_dataset(id=dataset_1_metadata.id)
            dataset_by_id_2 = await Actor.open_dataset(id=dataset_1_metadata.id)
            assert dataset_by_id_1 is dataset_by_name_1
            assert dataset_by_id_2 is dataset_by_id_1
        finally:
            await dataset_by_name_1.drop()


async def test_force_cloud(
    apify_client_async: ApifyClientAsync,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    """Test that force_cloud creates storage in the cloud even when running locally."""
    assert apify_client_async.token is not None
    monkeypatch.setenv(ApifyEnvVars.TOKEN, apify_client_async.token)

    dataset_name = generate_unique_resource_name('dataset')
    dataset_item = {'foo': 'bar'}

    async with Actor:
        dataset = await Actor.open_dataset(name=dataset_name, force_cloud=True)
        dataset_id = (await dataset.get_metadata()).id

        await dataset.push_data(dataset_item)

    dataset_client = apify_client_async.dataset(dataset_id)

    try:
        dataset_details = await dataset_client.get()
        assert dataset_details is not None
        assert dataset_details.get('name') == dataset_name

        dataset_items = await dataset_client.list_items()
        assert dataset_items.items == [dataset_item]
    finally:
        await dataset_client.delete()



from __future__ import annotations

from apify._configuration import Configuration
from apify.storage_clients._apify._alias_resolving import AliasResolver


def test_storage_key_format() -> None:
    """Test that _storage_key has the expected format: type,alias,hash."""
    config = Configuration(token='test-token', api_base_url='https://api.apify.com')
    resolver = AliasResolver(storage_type='Dataset', alias='my-alias', configuration=config)
    key = resolver._storage_key
    parts = key.split(',')
    assert len(parts) == 3
    assert parts[0] == 'Dataset'
    assert parts[1] == 'my-alias'
    assert len(parts[2]) > 0  # hash should be non-empty


async def test_resolve_id_returns_none_for_unknown() -> None:
    """Test that resolve_id returns None for an alias not in the map."""
    AliasResolver._alias_map = {}
    config = Configuration(token='test-token')
    resolver = AliasResolver(storage_type='Dataset', alias='unknown-alias', configuration=config)
    result = await resolver.resolve_id()
    assert result is None


async def test_resolve_id_returns_stored_id() -> None:
    """Test that resolve_id returns the ID if it was previously stored."""
    config = Configuration(token='test-token', api_base_url='https://api.apify.com')
    resolver = AliasResolver(storage_type='KeyValueStore', alias='test-alias', configuration=config)
    storage_key = resolver._storage_key
    AliasResolver._alias_map = {storage_key: 'stored-id-123'}

    result = await resolver.resolve_id()
    assert result == 'stored-id-123'


async def test_store_mapping_local_only() -> None:
    """Test that store_mapping only updates in-memory map when not at home."""
    AliasResolver._alias_map = {}
    config = Configuration(is_at_home=False, token='test-token')
    resolver = AliasResolver(storage_type='RequestQueue', alias='test-alias', configuration=config)

    await resolver.store_mapping(storage_id='new-id-456')

    # Verify the mapping was stored in memory
    assert AliasResolver._alias_map[resolver._storage_key] == 'new-id-456'


async def test_concurrent_alias_creation_uses_lock() -> None:
    """Test that the context manager acquires and releases a lock."""
    AliasResolver._alias_init_lock = None
    AliasResolver._alias_map = {}
    config = Configuration(token='test-token')
    resolver = AliasResolver(storage_type='Dataset', alias='test', configuration=config)

    async with resolver:
        # Lock should be acquired
        lock = await AliasResolver._get_alias_init_lock()
        assert lock.locked()

    # Lock should be released after exiting
    lock = await AliasResolver._get_alias_init_lock()
    assert not lock.locked()


async def test_get_alias_map_returns_in_memory_map() -> None:
    """Test that _get_alias_map returns the in-memory map when not at home."""
    AliasResolver._alias_map = {'existing_key': 'existing_id'}
    config = Configuration(is_at_home=False, token='test-token')

    result = await AliasResolver._get_alias_map(config)
    assert result == {'existing_key': 'existing_id'}
    # Also verify that an empty map is returned without fetching from KVS when not at home
    AliasResolver._alias_map = {}
    result = await AliasResolver._get_alias_map(config)
    assert result == {}


async def test_configuration_storages_alias_resolving() -> None:
    """Test that `AliasResolver` correctly resolves ids of storages registered in Configuration."""

    # Actor storages
    datasets = {'default': 'default_dataset_id', 'custom': 'custom_Dataset_id'}
    request_queues = {'default': 'default_request_queue_id', 'custom': 'custom_RequestQueue_id'}
    key_value_stores = {'default': 'default_key_value_store_id', 'custom': 'custom_KeyValueStore_id'}

    # Set up the configuration with the storage mapping
    configuration = Configuration(
        actor_storages={'datasets': datasets, 'request_queues': request_queues, 'key_value_stores': key_value_stores}
    )

    # Check that id of each non-default storage saved in the mapping is resolved
    for storage_type in ('Dataset', 'KeyValueStore', 'RequestQueue'):
        assert (
            await AliasResolver(storage_type=storage_type, alias='custom', configuration=configuration).resolve_id()
            == f'custom_{storage_type}_id'
        )



from __future__ import annotations

from urllib.parse import urljoin

from playwright.async_api import async_playwright

from apify import Actor, Request

# Note: To run this Actor locally, ensure that Playwright browsers are installed.
# Run `playwright install --with-deps` in the Actor's virtual environment to install them.
# When running on the Apify platform, these dependencies are already included
# in the Actor's Docker image.


async def main() -> None:
    # Enter the context of the Actor.
    async with Actor:
        # Retrieve the Actor input, and use default values if not provided.
        actor_input = await Actor.get_input() or {}
        start_urls = actor_input.get('start_urls', [{'url': 'https://apify.com'}])
        max_depth = actor_input.get('max_depth', 1)

        # Exit if no start URLs are provided.
        if not start_urls:
            Actor.log.info('No start URLs specified in actor input, exiting...')
            await Actor.exit()

        # Open the default request queue for handling URLs to be processed.
        request_queue = await Actor.open_request_queue()

        # Enqueue the start URLs with an initial crawl depth of 0.
        for start_url in start_urls:
            url = start_url.get('url')
            Actor.log.info(f'Enqueuing {url} ...')
            new_request = Request.from_url(url, user_data={'depth': 0})
            await request_queue.add_request(new_request)

        Actor.log.info('Launching Playwright...')

        # Launch Playwright and open a new browser context.
        async with async_playwright() as playwright:
            # Configure the browser to launch in headless mode as per Actor configuration.
            browser = await playwright.chromium.launch(
                headless=Actor.config.headless,
                args=['--disable-gpu'],
            )
            context = await browser.new_context()

            # Process the URLs from the request queue.
            while request := await request_queue.fetch_next_request():
                url = request.url

                if not isinstance(request.user_data['depth'], (str, int)):
                    raise TypeError('Request.depth is an unexpected type.')

                depth = int(request.user_data['depth'])
                Actor.log.info(f'Scraping {url} (depth={depth}) ...')

                try:
                    # Open a new page in the browser context and navigate to the URL.
                    page = await context.new_page()
                    await page.goto(url)

                    # If the current depth is less than max_depth, find nested links
                    # and enqueue them.
                    if depth < max_depth:
                        for link in await page.locator('a').all():
                            link_href = await link.get_attribute('href')
                            link_url = urljoin(url, link_href)

                            if link_url.startswith(('http://', 'https://')):
                                Actor.log.info(f'Enqueuing {link_url} ...')
                                new_request = Request.from_url(
                                    link_url,
                                    user_data={'depth': depth + 1},
                                )
                                await request_queue.add_request(new_request)

                    # Extract the desired data.
                    data = {
                        'url': url,
                        'title': await page.title(),
                    }

                    # Store the extracted data to the default dataset.
                    await Actor.push_data(data)

                except Exception:
                    Actor.log.exception(f'Cannot extract data from {url}.')

                finally:
                    await page.close()
                    # Mark the request as handled to ensure it is not processed again.
                    await request_queue.mark_request_as_handled(request)



from crawlee._utils.try_import import install_import_hook as _install_import_hook
from crawlee._utils.try_import import try_import as _try_import

_install_import_hook(__name__)

# The following imports use try_import to handle optional dependencies, as they may not always be available.

with _try_import(__name__, 'run_scrapy_actor'):
    from ._actor_runner import run_scrapy_actor

with _try_import(__name__, 'initialize_logging'):
    from ._logging_config import initialize_logging

with _try_import(__name__, 'to_apify_request', 'to_scrapy_request'):
    from .requests import to_apify_request, to_scrapy_request

with _try_import(__name__, 'ApifyScheduler'):
    from .scheduler import ApifyScheduler

with _try_import(__name__, 'apply_apify_settings', 'get_basic_auth_header'):
    from .utils import apply_apify_settings, get_basic_auth_header


__all__ = [
    'ApifyScheduler',
    'apply_apify_settings',
    'get_basic_auth_header',
    'initialize_logging',
    'run_scrapy_actor',
    'to_apify_request',
    'to_scrapy_request',
]



from __future__ import annotations

from scrapy.settings import Settings

from apify.scrapy.utils import apply_apify_settings


def test_overrides_scheduler() -> None:
    settings = Settings()
    new_settings = apply_apify_settings(settings=settings)

    assert new_settings.get('SCHEDULER') == 'apify.scrapy.scheduler.ApifyScheduler'


def test_updates_item_pipelines() -> None:
    settings = Settings(
        {
            'ITEM_PIPELINES': {
                'scrapy.pipelines.files.FilesPipeline': 1,
            }
        }
    )
    new_settings = apply_apify_settings(settings=settings)

    assert new_settings.get('ITEM_PIPELINES') == {
        'scrapy.pipelines.files.FilesPipeline': 1,
        'apify.scrapy.pipelines.ActorDatasetPushPipeline': 1000,
    }


def test_updates_downloader_middlewares() -> None:
    settings = Settings(
        {
            'DOWNLOADER_MIDDLEWARES': {
                'scrapy.downloadermiddlewares.robotstxt.RobotsTxtMiddleware': 123,
                'scrapy.downloadermiddlewares.httpproxy.HttpProxyMiddleware': 234,
                'scrapy.downloadermiddlewares.httpcompression.HttpCompressionMiddleware': 543,
            },
        }
    )
    new_settings = apply_apify_settings(settings=settings)

    assert new_settings.get('DOWNLOADER_MIDDLEWARES') == {
        'apify.scrapy.middlewares.ApifyHttpProxyMiddleware': 750,
        'scrapy.downloadermiddlewares.httpcompression.HttpCompressionMiddleware': 543,
        'scrapy.downloadermiddlewares.httpproxy.HttpProxyMiddleware': None,
        'scrapy.downloadermiddlewares.robotstxt.RobotsTxtMiddleware': 123,
    }


def test_adds_proxy_config() -> None:
    settings = Settings()
    new_settings = apply_apify_settings(settings=settings)
    assert new_settings.get('APIFY_PROXY_SETTINGS') is None

    settings = Settings()
    proxy_config = {'useApifyProxy': True, 'apifyProxyGroups': []}
    new_settings = apply_apify_settings(settings=settings, proxy_config=proxy_config)
    assert new_settings.get('APIFY_PROXY_SETTINGS') == {'useApifyProxy': True, 'apifyProxyGroups': []}
